# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2023 Larry Woestman <LarryWoestman2@gmail.com>          *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Path
import CAMTests.PathTestUtils as PathTestUtils
from importlib import reload
from Path.Post.scripts import linuxcnc_post as postprocessor

Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class TestLinuxCNCPost(PathTestUtils.PathTestBase):
    @classmethod
    def setUpClass(cls):
        """setUpClass()...
        This method is called upon instantiation of this test class.  Add code
        and objects here that are needed for the duration of the test() methods
        in this class.  In other words, set up the 'global' test environment
        here; use the `setUp()` method to set up a 'local' test environment.
        This method does not have access to the class `self` reference, but it
        is able to call static methods within this same class.
        """

        # Open existing FreeCAD document with test geometry
        FreeCAD.newDocument("Unnamed")

    @classmethod
    def tearDownClass(cls):
        """tearDownClass()...
        This method is called prior to destruction of this test class.  Add
        code and objects here that cleanup the test environment after the
        test() methods in this class have been executed.  This method does
        not have access to the class `self` reference.  This method is able
        to call static methods within this same class.
        """
        # Close geometry document without saving
        FreeCAD.closeDocument(FreeCAD.ActiveDocument.Name)

    # Setup and tear down methods called before and after each unit test
    def setUp(self):
        """setUp()...
        This method is called prior to each `test()` method.  Add code and
        objects here that are needed for multiple `test()` methods.
        """
        self.doc = FreeCAD.ActiveDocument
        self.con = FreeCAD.Console
        self.docobj = FreeCAD.ActiveDocument.addObject("Path::Feature", "testpath")
        reload(
            postprocessor
        )  # technical debt.  This shouldn't be necessary but here to bypass a bug

    def tearDown(self):
        """tearDown()...
        This method is called after each test() method. Add cleanup instructions here.
        Such cleanup instructions will likely undo those in the setUp() method.
        """
        FreeCAD.ActiveDocument.removeObject("testpath")

    def compare_sixth_line(self, path_string, expected, args, debug=False):
        """Perform a test with a single comparison to the sixth line of the output."""
        nl = "\n"
        if path_string:
            self.docobj.Path = Path.Path([Path.Command(path_string)])
        else:
            self.docobj.Path = Path.Path([])
        postables = [self.docobj]
        gcode = postprocessor.export(postables, "-", args)
        if debug:
            print(f"--------{nl}{gcode}--------{nl}")
        self.assertEqual(gcode.splitlines()[5], expected)

    def test000(self):
        """Test Output Generation.
        Empty path.  Produces only the preamble and postable.
        """

        self.docobj.Path = Path.Path([])
        postables = [self.docobj]

        # Test generating with header
        # Header contains a time stamp that messes up unit testing.
        # Only test length of result.
        args = "--no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        self.assertTrue(len(gcode.splitlines()) == 13)

        # Test without header
        expected = """(begin preamble)
G17 G54 G40 G49 G80 G90
G21
(begin operation: testpath)
(machine units: mm/min)
(finish operation: testpath)
(begin postamble)
M05
G17 G54 G90 G80 G40
M2
"""

        self.docobj.Path = Path.Path([])
        postables = [self.docobj]

        args = "--no-header --no-show-editor"
        # args = ("--no-header --no-comments --no-show-editor --precision=2")
        gcode = postprocessor.export(postables, "-", args)
        self.assertEqual(gcode, expected)

        # test without comments
        expected = """G17 G54 G40 G49 G80 G90
G21
M05
G17 G54 G90 G80 G40
M2
"""

        args = "--no-header --no-comments --no-show-editor"
        # args = ("--no-header --no-comments --no-show-editor --precision=2")
        gcode = postprocessor.export(postables, "-", args)
        self.assertEqual(gcode, expected)

    def test010(self):
        """Test command Generation.
        Test Precision
        Test imperial / inches
        """
        self.compare_sixth_line(
            "G0 X10 Y20 Z30", "G0 X10.000 Y20.000 Z30.000 ", "--no-header --no-show-editor"
        )
        self.compare_sixth_line(
            "G0 X10 Y20 Z30",
            "G0 X10.00 Y20.00 Z30.00 ",
            "--no-header --precision=2 --no-show-editor",
        )

    def test020(self):
        """
        Test Line Numbers
        """
        self.compare_sixth_line(
            "G0 X10 Y20 Z30",
            "N160  G0 X10.000 Y20.000 Z30.000 ",
            "--no-header --line-numbers --no-show-editor",
        )

    def test030(self):
        """
        Test Pre-amble
        """

        self.docobj.Path = Path.Path([])
        postables = [self.docobj]

        args = "--no-header --no-comments --preamble='G18 G55' --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        result = gcode.splitlines()[0]
        self.assertEqual(result, "G18 G55")

    def test040(self):
        """
        Test Post-amble
        """
        self.docobj.Path = Path.Path([])
        postables = [self.docobj]
        args = "--no-header --no-comments --postamble='G0 Z50\nM2' --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        result = gcode.splitlines()[-2]
        self.assertEqual(result, "G0 Z50")
        self.assertEqual(gcode.splitlines()[-1], "M2")

    def test050(self):
        """
        Test inches
        """

        c = Path.Command("G0 X10 Y20 Z30")
        self.docobj.Path = Path.Path([c])
        postables = [self.docobj]

        args = "--no-header --inches --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        self.assertEqual(gcode.splitlines()[2], "G20")

        result = gcode.splitlines()[5]
        expected = "G0 X0.3937 Y0.7874 Z1.1811 "
        self.assertEqual(result, expected)

        # Technical debt.   The following test fails.  Precision not working
        # with imperial units.

        # args = ("--no-header --inches --precision=2")
        # gcode = postprocessor.export(postables, "-", args)
        # result = gcode.splitlines()[5]
        # expected = "G0 X0.39 Y0.78 Z1.18 "
        # self.assertEqual(result, expected)

    def test060(self):
        """
        Test test modal
        Suppress the command name if the same as previous
        """
        c = Path.Command("G0 X10 Y20 Z30")
        c1 = Path.Command("G0 X10 Y30 Z30")

        self.docobj.Path = Path.Path([c, c1])
        postables = [self.docobj]

        args = "--no-header --modal --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        result = gcode.splitlines()[6]
        expected = "X10.000 Y30.000 Z30.000 "
        self.assertEqual(result, expected)

    def test070(self):
        """
        Test axis modal
        Suppress the axis coordinate if the same as previous
        """
        c = Path.Command("G0 X10 Y20 Z30")
        c1 = Path.Command("G0 X10 Y30 Z30")

        self.docobj.Path = Path.Path([c, c1])
        postables = [self.docobj]

        args = "--no-header --axis-modal --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        result = gcode.splitlines()[6]
        expected = "G0 Y30.000 "
        self.assertEqual(result, expected)

    def test080(self):
        """
        Test tool change
        """
        c = Path.Command("M6 T2")
        c2 = Path.Command("M3 S3000")
        self.docobj.Path = Path.Path([c, c2])
        postables = [self.docobj]

        args = "--no-header --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        self.assertEqual(gcode.splitlines()[5], "M5")
        self.assertEqual(gcode.splitlines()[6], "M6 T2 ")
        self.assertEqual(gcode.splitlines()[7], "G43 H2 ")
        self.assertEqual(gcode.splitlines()[8], "M3 S3000 ")

        # suppress TLO
        args = "--no-header --no-tlo --no-show-editor"
        gcode = postprocessor.export(postables, "-", args)
        self.assertEqual(gcode.splitlines()[7], "M3 S3000 ")

    def test090(self):
        """
        Test comment
        """
        self.compare_sixth_line("(comment)", "(comment) ", "--no-header --no-show-editor")

    def test100(self):
        """Test A, B, & C axis output for values between 0 and 90 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A40 B50 C60",
            "G1 X10.000 Y20.000 Z30.000 A40.000 B50.000 C60.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A40 B50 C60",
            "G1 X0.3937 Y0.7874 Z1.1811 A40.0000 B50.0000 C60.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test110(self):
        """Test A, B, & C axis output for 89 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A89 B89 C89",
            "G1 X10.000 Y20.000 Z30.000 A89.000 B89.000 C89.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A89 B89 C89",
            "G1 X0.3937 Y0.7874 Z1.1811 A89.0000 B89.0000 C89.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test120(self):
        """Test A, B, & C axis output for 90 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A90 B90 C90",
            "G1 X10.000 Y20.000 Z30.000 A90.000 B90.000 C90.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A90 B90 C90",
            "G1 X0.3937 Y0.7874 Z1.1811 A90.0000 B90.0000 C90.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test130(self):
        """Test A, B, & C axis output for 91 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A91 B91 C91",
            "G1 X10.000 Y20.000 Z30.000 A91.000 B91.000 C91.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A91 B91 C91",
            "G1 X0.3937 Y0.7874 Z1.1811 A91.0000 B91.0000 C91.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test140(self):
        """Test A, B, & C axis output for values between 90 and 180 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A100 B110 C120",
            "G1 X10.000 Y20.000 Z30.000 A100.000 B110.000 C120.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A100 B110 C120",
            "G1 X0.3937 Y0.7874 Z1.1811 A100.0000 B110.0000 C120.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test150(self):
        """Test A, B, & C axis output for values between 180 and 360 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A240 B250 C260",
            "G1 X10.000 Y20.000 Z30.000 A240.000 B250.000 C260.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A240 B250 C260",
            "G1 X0.3937 Y0.7874 Z1.1811 A240.0000 B250.0000 C260.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test160(self):
        """Test A, B, & C axis output for values greater than 360 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A440 B450 C460",
            "G1 X10.000 Y20.000 Z30.000 A440.000 B450.000 C460.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A440 B450 C460",
            "G1 X0.3937 Y0.7874 Z1.1811 A440.0000 B450.0000 C460.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test170(self):
        """Test A, B, & C axis output for values between 0 and -90 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-40 B-50 C-60",
            "G1 X10.000 Y20.000 Z30.000 A-40.000 B-50.000 C-60.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-40 B-50 C-60",
            "G1 X0.3937 Y0.7874 Z1.1811 A-40.0000 B-50.0000 C-60.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test180(self):
        """Test A, B, & C axis output for values between -90 and -180 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-100 B-110 C-120",
            "G1 X10.000 Y20.000 Z30.000 A-100.000 B-110.000 C-120.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-100 B-110 C-120",
            "G1 X0.3937 Y0.7874 Z1.1811 A-100.0000 B-110.0000 C-120.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test190(self):
        """Test A, B, & C axis output for values between -180 and -360 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-240 B-250 C-260",
            "G1 X10.000 Y20.000 Z30.000 A-240.000 B-250.000 C-260.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-240 B-250 C-260",
            "G1 X0.3937 Y0.7874 Z1.1811 A-240.0000 B-250.0000 C-260.0000 ",
            "--no-header --inches --no-show-editor",
        )

    def test200(self):
        """Test A, B, & C axis output for values below -360 degrees"""
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-440 B-450 C-460",
            "G1 X10.000 Y20.000 Z30.000 A-440.000 B-450.000 C-460.000 ",
            "--no-header --no-show-editor",
        )
        self.compare_sixth_line(
            "G1 X10 Y20 Z30 A-440 B-450 C-460",
            "G1 X0.3937 Y0.7874 Z1.1811 A-440.0000 B-450.0000 C-460.0000 ",
            "--no-header --inches --no-show-editor",
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Modules that define the workbench GuiCommands to perform graphical actions.

These GUI Commands or tools are called by buttons, menus, contextual menus,
toolbars, and other graphical widgets.
They are normally loaded by the workbench's `InitGui` module.
In this workbench these GUI tools are loaded by the `DraftTools`
module, which itself is loaded by `InitGui.DraftWorkbench.Initialize`.

This means that any script or workbench that wishes to use
this workbench's tools can import the individual modules,
or just `DraftTools`, to have access to them.

::

    import draftguitools.gui_rectangles
    import draftguitools.gui_circles
    import draftguitools.gui_arcs
    import DraftTools

These modules should not be imported in a console-only session,
as they are not intended to be used without the graphical interface (GUI).
They are also not generally suitable for scripting
as they normally require graphical input in the task panel or a selection
in the 3D view (`Gui.Selection`).

Most of these GUI tools require certain components of Draft to exist
like the `gui_snapper.Snapper`and the `DraftGui.DraftToolBar` classes.
These classes are normally installed in the global `Gui` namespace,
so they are accessible at all times.

::

    import DraftGui
    import draftguitools.gui_snapper
    Gui.draftToolBar = DraftGui.DraftToolBar()
    Gui.Snapper = draftguitools.gui_snapper.Snapper()

These classes can be imported and initialized individually
but it is easier to set them up just by importing `DraftTools`.
"""
## \defgroup draftguitools draftguitools
# \ingroup DRAFT
# \brief Modules that define the workbench GuiCommands to perform actions.



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to produce a mirrored object.

It just creates a `Part::Mirroring` object, and sets the appropriate
`Source` and `Normal` properties.
"""
## @package mirror
# \ingroup draftfunctions
# \brief Provides functions to produce a mirrored object.

## \addtogroup draftfunctions
# @{
import FreeCAD as App
import WorkingPlane

from draftutils import gui_utils
from draftutils import utils
from draftutils.messages import _err
from draftutils.translate import translate

if App.GuiUp:
    import FreeCADGui as Gui


def mirror(objlist, p1, p2):
    """Create a mirror object from the provided list and line.

    It creates a `Part::Mirroring` object from the given `objlist` using
    a plane that is defined by the two given points `p1` and `p2`,
    and the Draft working plane normal.

    Parameters
    ----------
    objlist: single object or a list of objects
        A single object or a list of objects.

    p1: Base::Vector3
        Point 1 of the mirror plane. It is also used as the `Placement.Base`
        of the resulting object.

    p2: Base::Vector3
        Point 2 of the mirror plane.

    Returns
    -------
    None
        If the operation fails.

    list
        List of `Part::Mirroring` objects, or a single one
        depending on the input `objlist`.

    To Do
    -----
    Implement a mirror tool specific to the workbench that does not
    just use `Part::Mirroring`. It should create a derived object,
    that is, it should work similar to `Draft.offset`.
    """

    if not objlist:
        _err(translate("draft", "No object given"))
        return

    if p1 == p2:
        _err(translate("draft", "The two points are coincident"))
        return

    if not isinstance(objlist, list):
        objlist = [objlist]

    norm = WorkingPlane.get_working_plane(update=False).axis
    pnorm = p2.sub(p1).cross(norm).normalize()

    result = []

    for obj in objlist:
        mir = App.ActiveDocument.addObject("Part::Mirroring", "Mirror")
        mir.Label = obj.Label + " (" + translate("draft", "mirrored") + ")"
        mir.Source = obj
        mir.Base = p1
        mir.Normal = pnorm
        gui_utils.format_object(mir, obj)
        result.append(mir)

    if len(result) == 1:
        result = result[0]
        gui_utils.select(result)

    return result


## @}



# ***************************************************************************
# *   Copyright (c) 2014 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore

    # s=os.path.dirname(__file__)
    # s=os.path.join(s,"ProfileLib")
    # FreeCADGui.addIconPath(s)

__title__ = "Sketcher profile lib handling"
__author__ = "Juergen Riegel"
__url__ = "https://www.freecad.org"


def isProfileActive():
    return not FreeCAD.ActiveDocument is None


def getSketch():
    edit = FreeCADGui.ActiveDocument.getInEdit()
    if edit and edit.isDerivedFrom("SketcherGui::ViewProviderSketch"):
        return edit.Object
    # act = FreeCAD.ActiveDocument.ActiveObject
    # if act and act.isDerivedFrom('Sketcher::SketchObject'):
    #    return act
    return None


class _CommandProfileHexagon1:
    "The basis hexagon profile command definition"

    def GetResources(self):
        return {
            "Pixmap": "Sketcher_CreateHexagon",
            "MenuText": QtCore.QT_TRANSLATE_NOOP(
                "Sketcher_ProfilesHexagon1", "Creates a hexagonal profile"
            ),
            "Accel": "",
            "CmdType": "ForEdit",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "Sketcher_ProfilesHexagon1", "Creates a hexagonal profile in the sketch"
            ),
        }

    def Activated(self):
        FreeCAD.ActiveDocument.openTransaction("Create hexagon profile")
        FreeCADGui.addModule("ProfileLib.Hexagon")
        sketch = getSketch()
        if not sketch is None:
            FreeCADGui.doCommand("ProfileLib.Hexagon.makeHexagonSimple('%s')" % (sketch.Name))
        else:
            FreeCADGui.doCommand("ProfileLib.Hexagon.makeHexagonSimple()")
        FreeCAD.ActiveDocument.recompute()

    def IsActive(self):
        return isProfileActive()


FreeCADGui.addCommand("Sketcher_ProfilesHexagon1", _CommandProfileHexagon1())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Modified 2016-01-03 JAndersM

__title__ = "FreeCAD Structure"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchStructure
#  \ingroup ARCH
#  \brief The Structure object and tools
#
#  This module provides tools to build Structure objects.
#  Structure elements are beams, columns, slabs, and other
#  elements that have a structural function, that is, that
#  support other parts of the building.

import FreeCAD
import ArchComponent
import ArchCommands
import ArchProfile
import Draft
import DraftVecUtils

from FreeCAD import Vector
from draftutils import params
from draftutils import gui_utils

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    import ArchPrecast
    import draftguitools.gui_trackers as DraftTrackers
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


# Reads preset profiles and categorizes them
Categories = []
Presets = ArchProfile.readPresets()
for pre in Presets:
    if pre[1] not in Categories:
        Categories.append(pre[1])


def makeStructure(baseobj=None, length=None, width=None, height=None, name=None):
    """makeStructure([baseobj],[length],[width],[height],[name]): creates a
    structure element based on the given profile object and the given
    extrusion height. If no base object is given, you can also specify
    length and width for a cubic object."""

    if not FreeCAD.ActiveDocument:
        FreeCAD.Console.PrintError("No active document. Aborting\n")
        return
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Structure")
    _Structure(obj)
    if FreeCAD.GuiUp:
        _ViewProviderStructure(obj.ViewObject)
    if baseobj:
        obj.Base = baseobj
        if FreeCAD.GuiUp:
            obj.Base.ViewObject.hide()
    if width:
        obj.Width = width
    else:
        obj.Width = params.get_param_arch("StructureWidth")
    if height:
        obj.Height = height
    else:
        if not length:
            obj.Height = params.get_param_arch("StructureHeight")
    if length:
        obj.Length = length
    else:
        if not baseobj:
            # don't set the length if we have a base object, otherwise the length X height calc
            # gets wrong
            obj.Length = params.get_param_arch("StructureLength")
    if baseobj:
        w = 0
        h = 0
        if hasattr(baseobj, "Width") and hasattr(baseobj, "Height"):
            w = baseobj.Width.Value
            h = baseobj.Height.Value
        elif hasattr(baseobj, "Length") and hasattr(baseobj, "Width"):
            w = baseobj.Length.Value
            h = baseobj.Width.Value
        elif hasattr(baseobj, "Length") and hasattr(baseobj, "Height"):
            w = baseobj.Length.Value
            h = baseobj.Height.Value
        if w and h:
            if length and not height:
                obj.Width = w
                obj.Height = h
            elif height and not length:
                obj.Width = w
                obj.Length = h

    if obj.Length > obj.Height:
        obj.IfcType = "Beam"
        obj.Label = name if name else translate("Arch", "Beam")
    elif obj.Height > obj.Length:
        obj.IfcType = "Column"
        obj.Label = name if name else translate("Arch", "Column")
    return obj


def makeStructuralSystem(objects=[], axes=[], name=None):
    """makeStructuralSystem([objects],[axes],[name]): makes a structural system
    based on the given objects and axes"""

    if not FreeCAD.ActiveDocument:
        FreeCAD.Console.PrintError("No active document. Aborting\n")
        return
    result = []
    if not axes:
        print("At least one axis must be given")
        return
    if objects:
        if not isinstance(objects, list):
            objects = [objects]
    else:
        objects = [None]
    for o in objects:
        obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "StructuralSystem")
        obj.Label = name if name else translate("Arch", "StructuralSystem")
        _StructuralSystem(obj)
        if FreeCAD.GuiUp:
            _ViewProviderStructuralSystem(obj.ViewObject)
        if o:
            obj.Base = o
        obj.Axes = axes
        result.append(obj)
        if FreeCAD.GuiUp and o:
            o.ViewObject.hide()
            Draft.formatObject(obj, o)
    FreeCAD.ActiveDocument.recompute()
    if len(result) == 1:
        return result[0]
    else:
        return result


def placeAlongEdge(p1, p2, horizontal=False):
    """placeAlongEdge(p1,p2,[horizontal]): returns a Placement positioned at p1, with Z axis oriented towards p2.
    If horizontal is True, then the X axis is oriented towards p2, not the Z axis"""

    pl = FreeCAD.Placement()
    pl.Base = p1
    import WorkingPlane

    up = WorkingPlane.get_working_plane(update=False).axis
    zaxis = p2.sub(p1)
    yaxis = up.cross(zaxis)
    if yaxis.Length > 0:
        xaxis = zaxis.cross(yaxis)
        if horizontal:
            pl.Rotation = FreeCAD.Rotation(zaxis, yaxis, xaxis, "ZXY")
        else:
            pl.Rotation = FreeCAD.Rotation(xaxis, yaxis, zaxis, "ZXY")
            pl.Rotation = FreeCAD.Rotation(
                pl.Rotation.multVec(FreeCAD.Vector(0, 0, 1)), 90
            ).multiply(pl.Rotation)
    return pl


class CommandStructuresFromSelection:
    """The Arch Structures from selection command definition."""

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Arch_MultipleStructures",
            "MenuText": QT_TRANSLATE_NOOP("Arch_StructuresFromSelection", "Multiple Structures"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_StructuresFromSelection",
                "Creates multiple BIM Structures from a selected base, using each selected edge as an extrusion path",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        selex = FreeCADGui.Selection.getSelectionEx()
        if len(selex) >= 2:
            FreeCAD.ActiveDocument.openTransaction(
                translate("Arch", "Create Structures From Selection")
            )
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            base = selex[
                0
            ].Object  # The first selected object is the base for the Structure objects
            for selexi in selex[
                1:
            ]:  # All the edges from the other objects are used as a Tool (extrusion paths)
                if len(selexi.SubElementNames) == 0:
                    subelement_names = [
                        "Edge" + str(i) for i in range(1, len(selexi.Object.Shape.Edges) + 1)
                    ]
                else:
                    subelement_names = [
                        sub for sub in selexi.SubElementNames if sub.startswith("Edge")
                    ]
                for sub in subelement_names:
                    FreeCADGui.doCommand(
                        "structure = Arch.makeStructure(FreeCAD.ActiveDocument." + base.Name + ")"
                    )
                    FreeCADGui.doCommand(
                        "structure.Tool = (FreeCAD.ActiveDocument."
                        + selexi.Object.Name
                        + ", '"
                        + sub
                        + "')"
                    )
                    FreeCADGui.doCommand("structure.BasePerpendicularToTool = True")
                    FreeCADGui.doCommand("Draft.autogroup(structure)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()
        else:
            FreeCAD.Console.PrintError(
                translate(
                    "Arch",
                    "Select the base object first and then the edges to use as extrusion paths",
                )
                + "\n"
            )


class CommandStructuralSystem:
    """The Arch Structural System command definition."""

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Arch_StructuralSystem",
            "MenuText": QT_TRANSLATE_NOOP("Arch_StructuralSystem", "Structural System"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_StructuralSystem",
                "Create a structural system from a selected structure and axis",
            ),
        }

    def IsActive(self):
        return not FreeCAD.ActiveDocument is None

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            st = Draft.getObjectsOfType(sel, "Structure")
            ax = Draft.getObjectsOfType(sel, "Axis")
            if ax:
                FreeCAD.ActiveDocument.openTransaction(
                    translate("Arch", "Create Structural System")
                )
                FreeCADGui.addModule("Arch")
                if st:
                    FreeCADGui.doCommand(
                        "obj = Arch.makeStructuralSystem("
                        + ArchCommands.getStringList(st)
                        + ", "
                        + ArchCommands.getStringList(ax)
                        + ")"
                    )
                else:
                    FreeCADGui.doCommand(
                        "obj = Arch.makeStructuralSystem(axes = "
                        + ArchCommands.getStringList(ax)
                        + ")"
                    )
                FreeCADGui.addModule("Draft")
                FreeCADGui.doCommand("Draft.autogroup(obj)")
                FreeCAD.ActiveDocument.commitTransaction()
                FreeCAD.ActiveDocument.recompute()
            else:
                FreeCAD.Console.PrintError(
                    translate("Arch", "Select at least an axis object") + "\n"
                )


class _CommandStructure:
    "the Arch Structure command definition"

    def __init__(self):

        self.beammode = False

    def GetResources(self):

        return {
            "Pixmap": "Arch_Structure",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Structure", "Structure"),
            "Accel": "S, T",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Structure",
                "Creates a structure from scratch or from a selected object (sketch, wire, face or solid)",
            ),
        }

    def IsActive(self):

        return not FreeCAD.ActiveDocument is None

    def Activated(self):

        self.doc = FreeCAD.ActiveDocument
        self.Width = params.get_param_arch("StructureWidth")
        if self.beammode:
            self.Height = params.get_param_arch("StructureLength")
            self.Length = params.get_param_arch("StructureHeight")
        else:
            self.Length = params.get_param_arch("StructureLength")
            self.Height = params.get_param_arch("StructureHeight")
        self.Profile = None
        self.bpoint = None
        self.bmode = False
        self.precastvalues = None
        self.wp = None
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            st = Draft.getObjectsOfType(sel, "Structure")
            ax = Draft.getObjectsOfType(sel, "Axis")
            if ax:
                FreeCADGui.runCommand("Arch_StructuralSystem")
                return
            elif not (ax) and not (st):
                self.doc.openTransaction(translate("Arch", "Create Structure"))
                FreeCADGui.addModule("Arch")
                for obj in sel:
                    FreeCADGui.doCommand(
                        "obj = Arch.makeStructure(FreeCAD.ActiveDocument." + obj.Name + ")"
                    )
                    FreeCADGui.addModule("Draft")
                    FreeCADGui.doCommand("Draft.autogroup(obj)")
                self.doc.commitTransaction()
                self.doc.recompute()
                return

        # interactive mode
        import WorkingPlane

        FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
        self.wp = WorkingPlane.get_working_plane()
        self.points = []
        self.tracker = DraftTrackers.boxTracker()
        self.tracker.width(self.Width)
        self.tracker.height(self.Height)
        self.tracker.length(self.Length)
        self.tracker.setRotation(self.wp.get_placement().Rotation)
        self.tracker.on()
        self.precast = ArchPrecast._PrecastTaskPanel()
        self.dents = ArchPrecast._DentsTaskPanel()
        self.precast.Dents = self.dents
        if self.beammode:
            title = translate("Arch", "First point of the beam")
        else:
            title = translate("Arch", "Base point of column")
        FreeCADGui.Snapper.getPoint(
            callback=self.getPoint,
            movecallback=self.update,
            extradlg=[self.taskbox(), self.precast.form, self.dents.form],
            title=title,
        )
        FreeCADGui.draftToolBar.continueCmd.show()

    def getPoint(self, point=None, obj=None):
        "this function is called by the snapper when it has a 3D point"

        self.bmode = self.modeb.isChecked()
        if point is None:
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.tracker.finalize()
            return
        if self.bmode and (self.bpoint is None):
            self.bpoint = point
            FreeCADGui.Snapper.getPoint(
                last=point,
                callback=self.getPoint,
                movecallback=self.update,
                extradlg=[self.taskbox(), self.precast.form, self.dents.form],
                title=translate("Arch", "Next point") + ":",
                mode="line",
            )
            return
        FreeCAD.activeDraftCommand = None
        FreeCADGui.Snapper.off()
        self.tracker.off()
        horiz = True  # determines the type of rotation to apply to the final object
        self.doc.openTransaction(translate("Arch", "Create Structure"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.addModule("WorkingPlane")
        if self.bmode:
            self.Length = point.sub(self.bpoint).Length
            params.set_param_arch("StructureHeight", self.Length)
        if self.Profile is not None:
            try:  # try to update latest precast values - fails if dialog has been destroyed already
                self.precastvalues = self.precast.getValues()
            except Exception:
                pass
            if ("Precast" in self.Profile) and self.precastvalues:
                # precast concrete
                self.precastvalues["PrecastType"] = self.Profile.split("_")[1]
                self.precastvalues["Length"] = self.Length
                self.precastvalues["Width"] = self.Width
                self.precastvalues["Height"] = self.Height
                argstring = ""
                # fix for precast placement, since their (0,0) point is the lower left corner
                if self.bmode:
                    delta = FreeCAD.Vector(0, 0 - self.Width / 2, 0)
                else:
                    delta = FreeCAD.Vector(-self.Length / 2, -self.Width / 2, 0)
                delta = self.wp.get_global_coords(delta, as_vector=True)
                point = point.add(delta)
                if self.bpoint:
                    self.bpoint = self.bpoint.add(delta)
                # build the string definition
                for pair in self.precastvalues.items():
                    argstring += pair[0].lower() + "="
                    if isinstance(pair[1], str):
                        argstring += '"' + pair[1] + '",'
                    else:
                        argstring += str(pair[1]) + ","
                FreeCADGui.addModule("ArchPrecast")
                FreeCADGui.doCommand("s = ArchPrecast.makePrecast(" + argstring + ")")
            else:
                # metal profile
                FreeCADGui.doCommand("p = Arch.makeProfile(" + str(self.Profile) + ")")
                if (
                    abs(self.Length - self.Profile[4]) >= 0.1
                ) or self.bmode:  # forgive rounding errors
                    # horizontal
                    FreeCADGui.doCommand(
                        "s = Arch.makeStructure(p,length=" + str(self.Length) + ")"
                    )
                    horiz = False
                else:
                    # vertical
                    FreeCADGui.doCommand(
                        "s = Arch.makeStructure(p,height=" + str(self.Height) + ")"
                    )
                    # if not self.bmode:
                    #    FreeCADGui.doCommand('s.Placement.Rotation = FreeCAD.Rotation(-0.5,0.5,-0.5,0.5)')
                FreeCADGui.doCommand('s.Profile = "' + self.Profile[2] + '"')
        else:
            FreeCADGui.doCommand(
                "s = Arch.makeStructure(length="
                + str(self.Length)
                + ",width="
                + str(self.Width)
                + ",height="
                + str(self.Height)
                + ")"
            )

        # calculate rotation
        if self.bmode and self.bpoint:
            FreeCADGui.doCommand(
                "s.Placement = Arch.placeAlongEdge("
                + DraftVecUtils.toString(self.bpoint)
                + ","
                + DraftVecUtils.toString(point)
                + ","
                + str(horiz)
                + ")"
            )
        else:
            FreeCADGui.doCommand("s.Placement.Base = " + DraftVecUtils.toString(point))
            FreeCADGui.doCommand("wp = WorkingPlane.get_working_plane()")
            FreeCADGui.doCommand(
                "s.Placement.Rotation = s.Placement.Rotation.multiply(wp.get_placement().Rotation)"
            )

        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("Draft.autogroup(s)")
        self.doc.commitTransaction()
        self.doc.recompute()
        # gui_utils.end_all_events()  # Causes a crash on Linux.
        self.tracker.finalize()
        if FreeCADGui.draftToolBar.continueMode:
            self.Activated()

    def _createItemlist(self, baselist):
        "create nice labels for presets in the task panel"

        ilist = []
        for p in baselist:
            f = FreeCAD.Units.Quantity(p[4], FreeCAD.Units.Length).getUserPreferred()
            d = params.get_param("Decimals", path="Units")
            s1 = str(round(p[4] / f[1], d))
            s2 = str(round(p[5] / f[1], d))
            s3 = str(f[2])
            ilist.append(p[2] + " (" + s1 + "x" + s2 + s3 + ")")
        return ilist

    def taskbox(self):
        "sets up a taskbox widget"

        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch", "Structure options"))
        grid = QtGui.QGridLayout(w)

        # mode box
        labelmode = QtGui.QLabel(translate("Arch", "Parameters of the structure") + ":")
        self.modeb = QtGui.QRadioButton(translate("Arch", "Beam"))
        self.modec = QtGui.QRadioButton(translate("Arch", "Column"))
        if self.bpoint or self.beammode:
            self.modeb.setChecked(True)
        else:
            self.modec.setChecked(True)
        grid.addWidget(labelmode, 0, 0, 1, 2)
        grid.addWidget(self.modeb, 1, 0, 1, 1)
        grid.addWidget(self.modec, 1, 1, 1, 1)

        # categories box
        labelc = QtGui.QLabel(translate("Arch", "Category"))
        self.valuec = QtGui.QComboBox()
        self.valuec.addItems([" ", "Precast concrete"] + Categories)
        grid.addWidget(labelc, 2, 0, 1, 1)
        grid.addWidget(self.valuec, 2, 1, 1, 1)

        # presets box
        labelp = QtGui.QLabel(translate("Arch", "Preset"))
        self.vPresets = QtGui.QComboBox()
        self.pSelect = [None]
        fpresets = [" "]
        self.vPresets.addItems(fpresets)
        grid.addWidget(labelp, 3, 0, 1, 1)
        grid.addWidget(self.vPresets, 3, 1, 1, 1)

        # length
        label1 = QtGui.QLabel(translate("Arch", "Length"))
        self.vLength = ui.createWidget("Gui::InputField")
        if self.modeb.isChecked():
            self.vLength.setText(
                FreeCAD.Units.Quantity(self.Height, FreeCAD.Units.Length).UserString
            )
        else:
            self.vLength.setText(
                FreeCAD.Units.Quantity(self.Length, FreeCAD.Units.Length).UserString
            )
        grid.addWidget(label1, 4, 0, 1, 1)
        grid.addWidget(self.vLength, 4, 1, 1, 1)

        # width
        label2 = QtGui.QLabel(translate("Arch", "Width"))
        self.vWidth = ui.createWidget("Gui::InputField")
        self.vWidth.setText(FreeCAD.Units.Quantity(self.Width, FreeCAD.Units.Length).UserString)
        grid.addWidget(label2, 5, 0, 1, 1)
        grid.addWidget(self.vWidth, 5, 1, 1, 1)

        # height
        label3 = QtGui.QLabel(translate("Arch", "Height"))
        self.vHeight = ui.createWidget("Gui::InputField")
        if self.modeb.isChecked():
            self.vHeight.setText(
                FreeCAD.Units.Quantity(self.Length, FreeCAD.Units.Length).UserString
            )
        else:
            self.vHeight.setText(
                FreeCAD.Units.Quantity(self.Height, FreeCAD.Units.Length).UserString
            )
        grid.addWidget(label3, 6, 0, 1, 1)
        grid.addWidget(self.vHeight, 6, 1, 1, 1)

        # horizontal button
        value4 = QtGui.QPushButton(translate("Arch", "Switch Length/Height"))
        grid.addWidget(value4, 7, 0, 1, 1)
        value5 = QtGui.QPushButton(translate("Arch", "Switch Length/Width"))
        grid.addWidget(value5, 7, 1, 1, 1)

        # connect slots
        QtCore.QObject.connect(
            self.valuec, QtCore.SIGNAL("currentIndexChanged(int)"), self.setCategory
        )
        QtCore.QObject.connect(
            self.vPresets, QtCore.SIGNAL("currentIndexChanged(int)"), self.setPreset
        )
        QtCore.QObject.connect(self.vLength, QtCore.SIGNAL("valueChanged(double)"), self.setLength)
        QtCore.QObject.connect(self.vWidth, QtCore.SIGNAL("valueChanged(double)"), self.setWidth)
        QtCore.QObject.connect(self.vHeight, QtCore.SIGNAL("valueChanged(double)"), self.setHeight)
        QtCore.QObject.connect(value4, QtCore.SIGNAL("pressed()"), self.rotateLH)
        QtCore.QObject.connect(value5, QtCore.SIGNAL("pressed()"), self.rotateLW)
        QtCore.QObject.connect(self.modeb, QtCore.SIGNAL("toggled(bool)"), self.switchLH)

        # restore preset
        stored = params.get_param_arch("StructurePreset")
        if stored:
            if stored.lower().startswith("precast_"):
                self.valuec.setCurrentIndex(1)
                tp = stored.split("_")[1]
                if tp and (tp in self.precast.PrecastTypes):
                    self.vPresets.setCurrentIndex(self.precast.PrecastTypes.index(tp))
            elif ";" in stored:
                stored = stored.split(";")
                if len(stored) >= 3:
                    if stored[1] in Categories:
                        self.valuec.setCurrentIndex(2 + Categories.index(stored[1]))
                        ps = [p[2] for p in self.pSelect]
                        if stored[2] in ps:
                            self.vPresets.setCurrentIndex(ps.index(stored[2]))
        return w

    def update(self, point, info):
        "this function is called by the Snapper when the mouse is moved"

        if FreeCADGui.Control.activeDialog():
            try:  # try to update latest precast values - fails if dialog has been destroyed already
                self.precastvalues = self.precast.getValues()
            except Exception:
                pass
            if self.Height >= self.Length:
                delta = Vector(0, 0, self.Height / 2)
            else:
                delta = Vector(self.Length / 2, 0, 0)
            delta = self.wp.get_global_coords(delta, as_vector=True)
            if self.modec.isChecked():
                self.tracker.pos(point.add(delta))
                self.tracker.on()
            else:
                if self.bpoint:
                    delta = Vector(0, 0, -self.Height / 2)
                    delta = self.wp.get_global_coords(delta, as_vector=True)
                    self.tracker.update([self.bpoint.add(delta), point.add(delta)])
                    self.tracker.on()
                    l = (point.sub(self.bpoint)).Length
                    self.vLength.setText(FreeCAD.Units.Quantity(l, FreeCAD.Units.Length).UserString)
                else:
                    self.tracker.off()

    def setWidth(self, d):

        self.Width = d
        self.tracker.width(d)
        params.set_param_arch("StructureWidth", d)

    def setHeight(self, d):

        self.Height = d
        self.tracker.height(d)
        if self.modeb.isChecked():
            params.set_param_arch("StructureLength", d)
        else:
            params.set_param_arch("StructureHeight", d)

    def setLength(self, d):

        self.Length = d
        self.tracker.length(d)
        if self.modeb.isChecked():
            params.set_param_arch("StructureHeight", d)
        else:
            params.set_param_arch("StructureLength", d)

    def setCategory(self, i):

        self.vPresets.clear()
        if i > 1:
            self.precast.form.hide()
            self.pSelect = [p for p in Presets if p[1] == Categories[i - 2]]
            fpresets = self._createItemlist(self.pSelect)
            self.vPresets.addItems(fpresets)
            self.setPreset(0)
        elif i == 1:
            self.precast.form.show()
            self.pSelect = self.precast.PrecastTypes
            fpresets = self.precast.PrecastTypes
            self.vPresets.addItems(fpresets)
            self.setPreset(0)
        else:
            self.precast.form.hide()
            self.pSelect = [None]
            fpresets = [" "]
            self.vPresets.addItems(fpresets)
            params.set_param_arch("StructurePreset", "")

    def setPreset(self, i):

        self.Profile = None
        elt = self.pSelect[i]
        if elt:
            if elt in self.precast.PrecastTypes:
                self.precast.setPreset(elt)
                self.Profile = "Precast_" + elt
                if elt in ["Pillar", "Beam"]:
                    self.dents.form.show()
                else:
                    self.dents.form.hide()
                params.set_param_arch("StructurePreset", self.Profile)
            else:
                self.vLength.setText(
                    FreeCAD.Units.Quantity(float(elt[4]), FreeCAD.Units.Length).UserString
                )
                self.vWidth.setText(
                    FreeCAD.Units.Quantity(float(elt[5]), FreeCAD.Units.Length).UserString
                )
                self.Profile = elt
                params.set_param_arch("StructurePreset", ";".join([str(i) for i in self.Profile]))

    def switchLH(self, bmode):

        if bmode:
            self.bmode = True
            if self.Height > self.Length:
                self.rotateLH()
        else:
            self.bmode = False
            if self.Length > self.Height:
                self.rotateLH()
                self.tracker.setRotation(FreeCAD.Rotation())

    def rotateLH(self):

        l = self.vLength.text()
        h = self.vHeight.text()
        self.vLength.setText(h)
        self.vHeight.setText(l)

    def rotateLW(self):

        l = self.vLength.text()
        w = self.vWidth.text()
        self.vLength.setText(w)
        self.vWidth.setText(l)


class _Structure(ArchComponent.Component):
    "The Structure object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Structure"
        self.setProperties(obj)
        obj.IfcType = "Beam"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Tool" in pl:
            obj.addProperty(
                "App::PropertyLinkSubList",
                "Tool",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP("App::Property", "An optional extrusion path for this element"),
                locked=True,
            )
        if not "ComputedLength" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "ComputedLength",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP("App::Property", "The computed length of the extrusion path"),
                1,
                locked=True,
            )
        if not "ToolOffsetFirst" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "ToolOffsetFirst",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Start offset distance along the extrusion path (positive: extend, negative: trim)",
                ),
                locked=True,
            )
        if not "ToolOffsetLast" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "ToolOffsetLast",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "End offset distance along the extrusion path (positive: extend, negative: trim)",
                ),
                locked=True,
            )
        if not "BasePerpendicularToTool" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "BasePerpendicularToTool",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Automatically align the Base of the Structure perpendicular to the Tool axis",
                ),
                locked=True,
            )
        if not "BaseOffsetX" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "BaseOffsetX",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "X offset between the Base origin and the Tool axis (only used if BasePerpendicularToTool is True)",
                ),
                locked=True,
            )
        if not "BaseOffsetY" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "BaseOffsetY",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Y offset between the Base origin and the Tool axis (only used if BasePerpendicularToTool is True)",
                ),
                locked=True,
            )
        if not "BaseMirror" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "BaseMirror",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Mirror the Base along its Y axis (only used if BasePerpendicularToTool is True)",
                ),
                locked=True,
            )
        if not "BaseRotation" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "BaseRotation",
                "ExtrusionPath",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Base rotation around the Tool axis (only used if BasePerpendicularToTool is True)",
                ),
                locked=True,
            )
        if not "Length" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Length",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The length of this element, if not based on a profile"
                ),
                locked=True,
            )
        if not "Width" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Width",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The width of this element, if not based on a profile"
                ),
                locked=True,
            )
        if not "Height" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The height or extrusion depth of this element. Keep 0 for automatic",
                ),
                locked=True,
            )
        if not "Normal" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "Normal",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The normal extrusion direction of this object (keep (0,0,0) for automatic normal)",
                ),
                locked=True,
            )
        if not "Nodes" in pl:
            obj.addProperty(
                "App::PropertyVectorList",
                "Nodes",
                "Structure",
                QT_TRANSLATE_NOOP("App::Property", "The structural nodes of this element"),
                locked=True,
            )
        if not "Profile" in pl:
            obj.addProperty(
                "App::PropertyString",
                "Profile",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "A description of the standard profile this element is based upon",
                ),
                locked=True,
            )
        if not "NodesOffset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "NodesOffset",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Offset distance between the centerline and the nodes line"
                ),
                locked=True,
            )
        if not "FaceMaker" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "FaceMaker",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The facemaker type to use to build the profile of this object"
                ),
                locked=True,
            )
            obj.FaceMaker = ["None", "Simple", "Cheese", "Bullseye"]
        if not "ArchSketchData" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "ArchSketchData",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Use Base ArchSketch (if used) data (e.g. widths, aligns, offsets) instead of Wall's properties",
                ),
                locked=True,
            )
            obj.ArchSketchData = True
        if not "ArchSketchEdges" in pl:  # PropertyStringList
            obj.addProperty(
                "App::PropertyStringList",
                "ArchSketchEdges",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Selected edges (or group of edges) of the base ArchSketch, to use in creating the shape of this BIM Structure (instead of using all the Base shape's edges by default).  Input are index numbers of edges or groups.",
                ),
                locked=True,
            )
        else:
            # test if the property was added but as IntegerList, then update;
            type = obj.getTypeIdOfProperty("ArchSketchEdges")
            if type == "App::PropertyIntegerList":
                oldIntValue = obj.ArchSketchEdges
                newStrValue = [str(x) for x in oldIntValue]
                obj.removeProperty("ArchSketchEdges")
                obj.addProperty(
                    "App::PropertyStringList",
                    "ArchSketchEdges",
                    "Structure",
                    QT_TRANSLATE_NOOP(
                        "App::Property",
                        "Selected edges (or group of edges) of the base ArchSketch, to use in creating the shape of this BIM Structure (instead of using all the Base shape's edges by default).  Input are index numbers of edges or groups.",
                    ),
                    locked=True,
                )
                obj.ArchSketchEdges = newStrValue
        if not hasattr(obj, "ArchSketchPropertySet"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "ArchSketchPropertySet",
                "Structure",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Select User Defined PropertySet to use in creating variant shape, with same ArchSketch ",
                ),
                locked=True,
            )
            obj.ArchSketchPropertySet = ["Default"]
        if not hasattr(self, "ArchSkPropSetPickedUuid"):
            self.ArchSkPropSetPickedUuid = ""
        if not hasattr(self, "ArchSkPropSetListPrev"):
            self.ArchSkPropSetListPrev = []

    def dumps(self):  # Supercede Arch.Component.dumps()
        dump = super().dumps()
        if not isinstance(dump, tuple):
            dump = (dump,)  # Python Tuple With One Item
        dump = dump + (self.ArchSkPropSetPickedUuid, self.ArchSkPropSetListPrev)
        return dump

    def loads(self, state):
        self.Type = "Structure"
        if state == None:
            return
        elif state[0] == "S":  # state[1] == 't', behaviour before 2024.11.28
            return
        elif state[0] == "Structure":
            self.ArchSkPropSetPickedUuid = state[1]
            self.ArchSkPropSetListPrev = state[2]
        elif state[0] != "Structure":  # model before merging super.dumps/loads()
            self.ArchSkPropSetPickedUuid = state[0]
            self.ArchSkPropSetListPrev = state[1]

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", ["ReadOnly"])
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", 0)
        else:
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", 0)
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", ["ReadOnly"])

        # set a flag to indicate onDocumentRestored() is run

    def execute(self, obj):
        "creates the structure shape"

        import Part
        import DraftGeomUtils

        if self.clone(obj):
            return
        if obj.Base and not self.ensureBase(obj):
            return

        base = None
        pl = obj.Placement

        # PropertySet support
        propSetPickedUuidPrev = self.ArchSkPropSetPickedUuid
        propSetListPrev = self.ArchSkPropSetListPrev
        propSetSelectedNamePrev = obj.ArchSketchPropertySet
        propSetSelectedNameCur = None
        propSetListCur = None
        if Draft.getType(obj.Base) == "ArchSketch":
            baseProxy = obj.Base.Proxy
            if hasattr(baseProxy, "getPropertySet"):
                # get full list of PropertySet
                propSetListCur = baseProxy.getPropertySet(obj.Base)
                # get updated name (if any) of the selected PropertySet
                propSetSelectedNameCur = baseProxy.getPropertySet(
                    obj.Base, propSetUuid=propSetPickedUuidPrev
                )
        if propSetSelectedNameCur:  # True if selection is not deleted
            if propSetListPrev != propSetListCur:
                obj.ArchSketchPropertySet = propSetListCur
                obj.ArchSketchPropertySet = propSetSelectedNameCur
                self.ArchSkPropSetListPrev = propSetListCur
            # elif propSetListPrev == propSetListCur:
            # pass  #nothing to do in this case
            # but if below, though (propSetListPrev == propSetListCur)
            elif propSetSelectedNamePrev != propSetSelectedNameCur:
                obj.ArchSketchPropertySet = propSetSelectedNameCur
        else:  # True if selection is deleted
            if propSetListCur:
                if propSetListPrev != propSetListCur:
                    obj.ArchSketchPropertySet = propSetListCur
                    obj.ArchSketchPropertySet = "Default"
                # else:  # Seems no need ...
                # obj.PropertySet = 'Default'

        extdata = self.getExtrusionData(obj)

        if extdata:
            sh = extdata[0]
            if not isinstance(sh, list):
                sh = [sh]
            ev = extdata[1]
            if not isinstance(ev, list):
                ev = [ev]
            pla = extdata[2]
            if not isinstance(pla, list):
                pla = [pla]
            base = []
            extrusion_length = 0.0
            for i in range(len(sh)):
                shi = sh[i]
                if i < len(ev):
                    evi = ev[i]
                else:
                    evi = ev[-1]
                    if isinstance(evi, FreeCAD.Vector):
                        evi = FreeCAD.Vector(evi)
                    else:
                        evi = evi.copy()
                if i < len(pla):
                    pli = pla[i]
                else:
                    pli = pla[-1].copy()
                shi.Placement = pli.multiply(shi.Placement)
                if isinstance(evi, FreeCAD.Vector):
                    extv = pla[0].Rotation.multVec(evi)
                    shi = shi.extrude(extv)
                else:
                    try:
                        shi = evi.makePipe(shi)
                    except Part.OCCError:
                        FreeCAD.Console.PrintError(
                            translate(
                                "Arch",
                                "Error: The base shape could not be extruded along this tool object",
                            )
                            + "\n"
                        )
                        return
                base.append(shi)
                extrusion_length += evi.Length
            if len(base) == 1:
                base = base[0]
            else:
                base = Part.makeCompound(base)
            obj.ComputedLength = FreeCAD.Units.Quantity(extrusion_length, FreeCAD.Units.Length)
        if obj.Base:
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape.isNull():
                    return
                if not obj.Base.Shape.isValid():
                    if not obj.Base.Shape.Solids:
                        # let pass invalid objects if they have solids...
                        return
                elif obj.Base.Shape.Solids:
                    base = obj.Base.Shape.copy()
            elif obj.Base.isDerivedFrom("Mesh::Feature"):
                if obj.Base.Mesh.isSolid():
                    if obj.Base.Mesh.countComponents() == 1:
                        sh = ArchCommands.getShapeFromMesh(obj.Base.Mesh)
                        if sh.isClosed() and sh.isValid() and sh.Solids and (not sh.isNull()):
                            base = sh
                        else:
                            FreeCAD.Console.PrintWarning(
                                translate("Arch", "This mesh is an invalid solid") + "\n"
                            )
                            obj.Base.ViewObject.show()
        if (not base) and (not obj.Additions):
            # FreeCAD.Console.PrintError(translate("Arch","Error: Invalid base object")+"\n")
            return

        base = self.processSubShapes(obj, base, pl)
        self.applyShape(obj, base, pl)

    def getExtrusionData(self, obj):
        """returns (shape,extrusion vector or path,placement) or None"""
        if hasattr(obj, "IfcType"):
            IfcType = obj.IfcType
        else:
            IfcType = None
        import Part
        import DraftGeomUtils

        data = ArchComponent.Component.getExtrusionData(self, obj)
        if data:
            if not isinstance(data[0], list):
                # multifuses not considered here
                return data
        length = obj.Length.Value
        width = obj.Width.Value
        height = obj.Height.Value
        if not height:
            height = self.getParentHeight(obj)
        baseface = None
        extrusion = None
        normal = None
        if obj.Base:
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape:
                    if obj.Base.Shape.Solids:
                        return None
                    elif obj.Base.Shape.Faces:
                        if not DraftGeomUtils.isCoplanar(obj.Base.Shape.Faces, tol=0.01):
                            return None
                        else:
                            baseface = obj.Base.Shape.copy()
                    elif obj.Base.Shape.Wires:
                        # ArchSketch feature :
                        # Get base shape wires, and faceMaker, for Structure (slab. etc.) from Base Objects if they store and provide by getStructureBaseShapeWires()
                        # (thickness, normal/extrusion, length, width, baseface maybe for later) of structure (slab etc.)
                        structureBaseShapeWires = []
                        baseShapeWires = []  # baseSlabWires / baseSlabOpeningWires = None
                        faceMaker = None

                        if (
                            hasattr(obj.Base, "Proxy")
                            and obj.ArchSketchData
                            and hasattr(obj.Base.Proxy, "getStructureBaseShapeWires")
                        ):
                            propSetUuid = self.ArchSkPropSetPickedUuid

                            # provide selected edges, or groups, in obj.ArchSketchEdges for processing in getStructureBaseShapeWires() (getSortedClusters) as override
                            structureBaseShapeWires = obj.Base.Proxy.getStructureBaseShapeWires(
                                obj.Base, propSetUuid=propSetUuid
                            )
                            # get slab wires; use original wires if structureBaseShapeWires() provided none
                            if (
                                structureBaseShapeWires
                            ):  # would be false (none) if both base ArchSketch and obj do not have the edges stored / inputted by user
                                # if structureBaseShapeWires is {dict}
                                baseShapeWires = structureBaseShapeWires.get("slabWires")
                                faceMaker = structureBaseShapeWires.get("faceMaker")
                        elif obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                            skGeom = obj.Base.GeometryFacadeList
                            skGeomEdges = []
                            skPlacement = (
                                obj.Base.Placement
                            )  # Get Sketch's placement to restore later
                            # Get ArchSketch edges to construct ArchStructure
                            # No need to test obj.ArchSketchData ...
                            for ig, geom in enumerate(skGeom):
                                # Construction mode edges should be ignored if
                                # ArchSketchEdges, otherwise, ArchSketchEdges data
                                # needs to take out those in Construction before
                                # using as parameters.
                                if (not obj.ArchSketchEdges and not geom.Construction) or str(
                                    ig
                                ) in obj.ArchSketchEdges:
                                    # support Line, Arc, Circle, Ellipse, BSplineCurve for Sketch
                                    # as Base at the moment
                                    if isinstance(
                                        geom.Geometry,
                                        (
                                            Part.LineSegment,
                                            Part.Circle,
                                            Part.ArcOfCircle,
                                            Part.Ellipse,
                                            Part.BSplineCurve,
                                        ),
                                    ):
                                        skGeomEdgesI = geom.Geometry.toShape()
                                        skGeomEdges.append(skGeomEdgesI)
                            clusterTransformed = []
                            for cluster in Part.getSortedClusters(skGeomEdges):
                                edgesTransformed = []
                                for edge in cluster:
                                    edge.Placement = edge.Placement.multiply(skPlacement)
                                    edgesTransformed.append(edge)
                                clusterTransformed.append(edgesTransformed)
                            for clusterT in clusterTransformed:
                                baseShapeWires.append(Part.Wire(clusterT))

                        if not baseShapeWires:
                            baseShapeWires = obj.Base.Shape.Wires
                        if faceMaker or (obj.FaceMaker != "None"):
                            if not faceMaker:
                                faceMaker = obj.FaceMaker
                            try:
                                baseface = Part.makeFace(
                                    baseShapeWires, "Part::FaceMaker" + str(faceMaker)
                                )
                            except Exception:
                                FreeCAD.Console.PrintError(
                                    translate("Arch", "Facemaker returned an error") + "\n"
                                )
                                # Not returning even Part.makeFace fails, fall back to 'non-Part.makeFace' method
                        if not baseface:
                            for w in baseShapeWires:
                                if not w.isClosed():
                                    p0 = w.OrderedVertexes[0].Point
                                    p1 = w.OrderedVertexes[-1].Point
                                    if p0 != p1:
                                        e = Part.LineSegment(p0, p1).toShape()
                                        w.add(e)
                                w.fix(0.1, 0, 1)  # fixes self-intersecting wires
                                f = Part.Face(w)
                                # check if it is 1st face (f) created from w in baseShapeWires; if not, fuse()
                                if baseface:
                                    baseface = baseface.fuse(f)
                                else:
                                    # TODO use Part.Shape() rather than shape.copy() ... ?
                                    baseface = f.copy()
        elif length and width and height:
            # check if this was a profil based arch structure
            # profile-based structures should use XY plane orientation
            use_profile_orientation = hasattr(obj, "Profile") and obj.Profile

            if (length > height) and (IfcType in ["Beam", "Column"]):
                h2 = height / 2 or 0.5
                w2 = width / 2 or 0.5
                if use_profile_orientation:
                    v1 = Vector(-w2, -h2, 0)
                    v2 = Vector(w2, -h2, 0)
                    v3 = Vector(w2, h2, 0)
                    v4 = Vector(-w2, h2, 0)
                else:
                    v1 = Vector(0, -w2, -h2)
                    v2 = Vector(0, w2, -h2)
                    v3 = Vector(0, w2, h2)
                    v4 = Vector(0, -w2, h2)
            else:
                l2 = length / 2 or 0.5
                w2 = width / 2 or 0.5
                v1 = Vector(-l2, -w2, 0)
                v2 = Vector(l2, -w2, 0)
                v3 = Vector(l2, w2, 0)
                v4 = Vector(-l2, w2, 0)
            import Part

            baseface = Part.Face(Part.makePolygon([v1, v2, v3, v4, v1]))
        if baseface:
            if hasattr(obj, "Tool") and obj.Tool:
                tool = obj.Tool
                edges = DraftGeomUtils.get_referenced_edges(tool)
                if len(edges) > 0:
                    extrusion = Part.Wire(Part.__sortEdges__(edges))
                    if hasattr(obj, "ToolOffsetFirst"):
                        offset_start = float(obj.ToolOffsetFirst.getValueAs("mm"))
                    else:
                        offset_start = 0.0
                    if hasattr(obj, "ToolOffsetLast"):
                        offset_end = float(obj.ToolOffsetLast.getValueAs("mm"))
                    else:
                        offset_end = 0.0
                    if offset_start != 0.0 or offset_end != 0.0:
                        extrusion = DraftGeomUtils.get_extended_wire(
                            extrusion, offset_start, offset_end
                        )
                    if hasattr(obj, "BasePerpendicularToTool") and obj.BasePerpendicularToTool:
                        pl = FreeCAD.Placement()
                        if hasattr(obj, "BaseRotation"):
                            pl.rotate(
                                FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 0, 1), -obj.BaseRotation
                            )
                        if hasattr(obj, "BaseOffsetX") and hasattr(obj, "BaseOffsetY"):
                            pl.translate(FreeCAD.Vector(obj.BaseOffsetX, obj.BaseOffsetY, 0))
                        if hasattr(obj, "BaseMirror") and obj.BaseMirror:
                            pl.rotate(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 1, 0), 180)
                        baseface.Placement = DraftGeomUtils.get_placement_perpendicular_to_wire(
                            extrusion
                        ).multiply(pl)
            else:
                if obj.Normal.Length:
                    normal = Vector(obj.Normal).normalize()
                else:
                    normal = baseface.Faces[0].normalAt(
                        0, 0
                    )  # TODO to use ArchSketch's 'normal' for consistency
            base = None
            placement = None
            inverse_placement = None
            if len(baseface.Faces) > 1:
                base = []
                placement = []
                hint = baseface.Faces[0].normalAt(0, 0)  # TODO anything to do ?
                for f in baseface.Faces:
                    bf, pf = self.rebase(f, hint)
                    base.append(bf)
                    placement.append(pf)
                inverse_placement = placement[0].inverse()
            else:
                base, placement = self.rebase(baseface)
                inverse_placement = placement.inverse()
            if extrusion:
                if (
                    len(extrusion.Edges) == 1
                    and DraftGeomUtils.geomType(extrusion.Edges[0]) == "Line"
                ):
                    extrusion = DraftGeomUtils.vec(extrusion.Edges[0], True)
                if isinstance(extrusion, FreeCAD.Vector):
                    extrusion = inverse_placement.Rotation.multVec(extrusion)
            elif normal:
                normal = inverse_placement.Rotation.multVec(normal)
                if not normal:
                    normal = Vector(0, 0, 1)
                if not normal.Length:
                    normal = Vector(0, 0, 1)
                extrusion = normal
                if (length > height) and (IfcType in ["Beam", "Column"]):
                    if length:
                        extrusion = normal.multiply(length)
                else:
                    if height:
                        extrusion = normal.multiply(height)
            if extrusion:
                return (base, extrusion, placement)
        return None

    def onChanged(self, obj, prop):

        # check the flag indicating if we are currently in the process of
        # restoring document; if not, no further code is run as getExtrusionData()
        # below return error when some properties are not added by onDocumentRestored()
        if FreeCAD.ActiveDocument.Restoring:
            return

        if hasattr(obj, "IfcType"):
            IfcType = obj.IfcType
        else:
            IfcType = None
        self.hideSubobjects(obj, prop)
        if prop in ["Shape", "ResetNodes", "NodesOffset"]:
            # ResetNodes is not a property but it allows us to use this function to force reset the nodes
            nodes = None
            extdata = self.getExtrusionData(obj)
            if extdata and not isinstance(extdata[0], list):
                nodes = extdata[0]
                if IfcType in ["Beam", "Column"]:
                    if not isinstance(extdata[1], FreeCAD.Vector):
                        nodes = extdata[1]
                    elif extdata[1].Length > 0:
                        if hasattr(nodes, "CenterOfMass"):
                            import Part

                            nodes = Part.LineSegment(
                                nodes.CenterOfMass, nodes.CenterOfMass.add(extdata[1])
                            ).toShape()
                if isinstance(extdata[1], FreeCAD.Vector):
                    nodes.Placement = nodes.Placement.multiply(extdata[2])
            offset = FreeCAD.Vector()
            if hasattr(obj, "NodesOffset"):
                offset = FreeCAD.Vector(0, 0, obj.NodesOffset.Value)
            if obj.Nodes and (prop != "ResetNodes"):
                if hasattr(self, "nodes"):
                    if self.nodes:
                        if obj.Nodes != self.nodes:
                            # nodes are set manually: don't touch them
                            return
                else:
                    # nodes haven't been calculated yet, but are set (file load)
                    # we set the nodes now but don't change the property
                    if nodes:
                        self.nodes = [v.Point.add(offset) for v in nodes.Vertexes]
                        return
            # we set the nodes
            if nodes:
                self.nodes = [v.Point.add(offset) for v in nodes.Vertexes]
                obj.Nodes = self.nodes
        ArchComponent.Component.onChanged(self, obj, prop)

        if prop == "ArchSketchPropertySet" and Draft.getType(obj.Base) == "ArchSketch":
            baseProxy = obj.Base.Proxy
            if hasattr(baseProxy, "getPropertySet"):
                uuid = baseProxy.getPropertySet(obj, propSetName=obj.ArchSketchPropertySet)
                self.ArchSkPropSetPickedUuid = uuid
        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", ["ReadOnly"])
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", 0)
        else:
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", 0)
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", ["ReadOnly"])

    def getNodeEdges(self, obj):
        "returns a list of edges from structural nodes"

        edges = []
        if obj.Nodes:
            import Part

            for i in range(len(obj.Nodes) - 1):
                edges.append(
                    Part.LineSegment(
                        obj.Placement.multVec(obj.Nodes[i]), obj.Placement.multVec(obj.Nodes[i + 1])
                    ).toShape()
                )
            if hasattr(obj.ViewObject, "NodeType"):
                if (obj.ViewObject.NodeType == "Area") and (len(obj.Nodes) > 2):
                    edges.append(
                        Part.LineSegment(
                            obj.Placement.multVec(obj.Nodes[-1]),
                            obj.Placement.multVec(obj.Nodes[0]),
                        ).toShape()
                    )
        return edges


class _ViewProviderStructure(ArchComponent.ViewProviderComponent):
    "A View Provider for the Structure object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

        # setProperties of ArchComponent will be overwritten
        # thus setProperties from ArchComponent will be explicit called to get the properties
        ArchComponent.ViewProviderComponent.setProperties(self, vobj)

        self.setProperties(vobj)
        vobj.ShapeColor = ArchCommands.getDefaultColor("Structure")

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "ShowNodes" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowNodes",
                "Nodes",
                QT_TRANSLATE_NOOP("App::Property", "If the nodes are visible or not"),
                locked=True,
            ).ShowNodes = False
        if not "NodeLine" in pl:
            vobj.addProperty(
                "App::PropertyFloat",
                "NodeLine",
                "Nodes",
                QT_TRANSLATE_NOOP("App::Property", "The width of the nodes line"),
                locked=True,
            )
        if not "NodeSize" in pl:
            vobj.addProperty(
                "App::PropertyFloat",
                "NodeSize",
                "Nodes",
                QT_TRANSLATE_NOOP("App::Property", "The size of the node points"),
                locked=True,
            )
            vobj.NodeSize = 6
        if not "NodeColor" in pl:
            vobj.addProperty(
                "App::PropertyColor",
                "NodeColor",
                "Nodes",
                QT_TRANSLATE_NOOP("App::Property", "The color of the nodes line"),
                locked=True,
            )
            vobj.NodeColor = (1.0, 1.0, 1.0, 1.0)
        if not "NodeType" in pl:
            vobj.addProperty(
                "App::PropertyEnumeration",
                "NodeType",
                "Nodes",
                QT_TRANSLATE_NOOP("App::Property", "The type of structural node"),
                locked=True,
            )
            vobj.NodeType = ["Linear", "Area"]

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        import Arch_rc

        if hasattr(self, "Object"):
            if hasattr(self.Object, "CloneOf"):
                if self.Object.CloneOf:
                    return ":/icons/Arch_Structure_Clone.svg"
        return ":/icons/Arch_Structure_Tree.svg"

    def updateData(self, obj, prop):

        if prop == "Nodes":
            if obj.Nodes:
                if hasattr(self, "nodes"):
                    p = []
                    self.pointset.numPoints.setValue(0)
                    self.lineset.coordIndex.deleteValues(0)
                    self.faceset.coordIndex.deleteValues(0)
                    for n in obj.Nodes:
                        p.append([n.x, n.y, n.z])
                    self.coords.point.setValues(0, len(p), p)
                    self.pointset.numPoints.setValue(len(p))
                    self.lineset.coordIndex.setValues(0, len(p) + 1, list(range(len(p))) + [-1])
                    if hasattr(obj.ViewObject, "NodeType"):
                        if (obj.ViewObject.NodeType == "Area") and (len(p) > 2):
                            self.coords.point.set1Value(len(p), p[0][0], p[0][1], p[0][2])
                            self.lineset.coordIndex.setValues(
                                0, len(p) + 2, list(range(len(p) + 1)) + [-1]
                            )
                            self.faceset.coordIndex.setValues(
                                0, len(p) + 1, list(range(len(p))) + [-1]
                            )

        elif prop in ["IfcType"]:
            if hasattr(obj.ViewObject, "NodeType"):
                if hasattr(obj, "IfcType"):
                    IfcType = obj.IfcType
                else:
                    IfcType = None
                if IfcType == "Slab":
                    obj.ViewObject.NodeType = "Area"
                else:
                    obj.ViewObject.NodeType = "Linear"
        else:
            ArchComponent.ViewProviderComponent.updateData(self, obj, prop)

    def onChanged(self, vobj, prop):

        if prop == "ShowNodes":
            if hasattr(self, "nodes"):
                vobj.Annotation.removeChild(self.nodes)
                del self.nodes
            if vobj.ShowNodes:
                from pivy import coin

                self.nodes = coin.SoAnnotation()
                self.coords = coin.SoCoordinate3()
                self.mat = coin.SoMaterial()
                self.pointstyle = coin.SoDrawStyle()
                self.pointstyle.style = coin.SoDrawStyle.POINTS
                self.pointset = coin.SoType.fromName("SoBrepPointSet").createInstance()
                self.linestyle = coin.SoDrawStyle()
                self.linestyle.style = coin.SoDrawStyle.LINES
                self.lineset = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
                self.facestyle = coin.SoDrawStyle()
                self.facestyle.style = coin.SoDrawStyle.FILLED
                self.shapehints = coin.SoShapeHints()
                self.shapehints.faceType = coin.SoShapeHints.UNKNOWN_FACE_TYPE
                self.fmat = coin.SoMaterial()
                self.fmat.transparency.setValue(0.75)
                self.faceset = coin.SoIndexedFaceSet()
                self.nodes.addChild(self.coords)
                self.nodes.addChild(self.mat)
                self.nodes.addChild(self.pointstyle)
                self.nodes.addChild(self.pointset)
                self.nodes.addChild(self.linestyle)
                self.nodes.addChild(self.lineset)
                self.nodes.addChild(self.facestyle)
                self.nodes.addChild(self.shapehints)
                self.nodes.addChild(self.fmat)
                self.nodes.addChild(self.faceset)
                vobj.Annotation.addChild(self.nodes)
                self.updateData(vobj.Object, "Nodes")
                self.onChanged(vobj, "NodeColor")
                self.onChanged(vobj, "NodeLine")
                self.onChanged(vobj, "NodeSize")

        elif prop == "NodeColor":
            if hasattr(self, "mat"):
                l = vobj.NodeColor
                self.mat.diffuseColor.setValue([l[0], l[1], l[2]])
                self.fmat.diffuseColor.setValue([l[0], l[1], l[2]])

        elif prop == "NodeLine":
            if hasattr(self, "linestyle"):
                self.linestyle.lineWidth = vobj.NodeLine

        elif prop == "NodeSize":
            if hasattr(self, "pointstyle"):
                self.pointstyle.pointSize = vobj.NodeSize

        elif prop == "NodeType":
            self.updateData(vobj.Object, "Nodes")

        else:
            ArchComponent.ViewProviderComponent.onChanged(self, vobj, prop)

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = StructureTaskPanel(vobj.Object)
        taskd.obj = self.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True


class StructureTaskPanel(ArchComponent.ComponentTaskPanel):

    def __init__(self, obj):

        ArchComponent.ComponentTaskPanel.__init__(self)
        self.nodes_widget = QtGui.QWidget()
        self.nodes_widget.setWindowTitle(QtGui.QApplication.translate("Arch", "Node Tools", None))
        lay = QtGui.QVBoxLayout(self.nodes_widget)

        self.resetButton = QtGui.QPushButton(self.nodes_widget)
        self.resetButton.setIcon(QtGui.QIcon(":/icons/edit-undo.svg"))
        self.resetButton.setText(QtGui.QApplication.translate("Arch", "Reset Nodes", None))

        lay.addWidget(self.resetButton)
        QtCore.QObject.connect(self.resetButton, QtCore.SIGNAL("clicked()"), self.resetNodes)

        self.editButton = QtGui.QPushButton(self.nodes_widget)
        self.editButton.setIcon(QtGui.QIcon(":/icons/Draft_Edit.svg"))
        self.editButton.setText(QtGui.QApplication.translate("Arch", "Edit Nodes", None))
        lay.addWidget(self.editButton)
        QtCore.QObject.connect(self.editButton, QtCore.SIGNAL("clicked()"), self.editNodes)

        self.extendButton = QtGui.QPushButton(self.nodes_widget)
        self.extendButton.setIcon(QtGui.QIcon(":/icons/Snap_Perpendicular.svg"))
        self.extendButton.setText(QtGui.QApplication.translate("Arch", "Extend Nodes", None))
        self.extendButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch",
                "Extends the nodes of this element to reach the nodes of another element",
                None,
            )
        )
        lay.addWidget(self.extendButton)
        QtCore.QObject.connect(self.extendButton, QtCore.SIGNAL("clicked()"), self.extendNodes)

        self.connectButton = QtGui.QPushButton(self.nodes_widget)
        self.connectButton.setIcon(QtGui.QIcon(":/icons/Snap_Intersection.svg"))
        self.connectButton.setText(QtGui.QApplication.translate("Arch", "Connect Nodes", None))
        self.connectButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Connects nodes of this element with the nodes of another element", None
            )
        )
        lay.addWidget(self.connectButton)
        QtCore.QObject.connect(self.connectButton, QtCore.SIGNAL("clicked()"), self.connectNodes)

        self.toggleButton = QtGui.QPushButton(self.nodes_widget)
        self.toggleButton.setIcon(QtGui.QIcon(":/icons/dagViewVisible.svg"))
        self.toggleButton.setText(QtGui.QApplication.translate("Arch", "Toggle All Nodes", None))
        self.toggleButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Toggles all structural nodes of the document on/off", None
            )
        )
        lay.addWidget(self.toggleButton)
        QtCore.QObject.connect(self.toggleButton, QtCore.SIGNAL("clicked()"), self.toggleNodes)

        self.extrusion_widget = QtGui.QWidget()
        self.extrusion_widget.setWindowTitle(
            QtGui.QApplication.translate("Arch", "Extrusion Tools", None)
        )
        lay = QtGui.QVBoxLayout(self.extrusion_widget)

        self.selectToolButton = QtGui.QPushButton(self.extrusion_widget)
        self.selectToolButton.setIcon(QtGui.QIcon())
        self.selectToolButton.setText(QtGui.QApplication.translate("Arch", "Select Tool", None))
        self.selectToolButton.setToolTip(
            QtGui.QApplication.translate(
                "Arch", "Selects object or edges to be used as a tool (extrusion path)", None
            )
        )
        lay.addWidget(self.selectToolButton)
        QtCore.QObject.connect(
            self.selectToolButton, QtCore.SIGNAL("clicked()"), self.setSelectionFromTool
        )

        self.form = [self.form, self.nodes_widget, self.extrusion_widget]
        self.Object = obj
        self.observer = None
        self.nodevis = None

    def editNodes(self):

        FreeCADGui.Control.closeDialog()
        FreeCADGui.runCommand("Draft_Edit")

    def resetNodes(self):

        self.Object.Proxy.onChanged(self.Object, "ResetNodes")

    def extendNodes(self, other=None):

        if not other:
            self.observer = StructSelectionObserver(self.extendNodes)
            FreeCADGui.Selection.addObserver(self.observer)
            FreeCAD.Console.PrintMessage(translate("Arch", "Choose another Structure object:"))
        else:
            FreeCADGui.Selection.removeObserver(self.observer)
            self.observer = None
            if Draft.getType(other) != "Structure":
                FreeCAD.Console.PrintError(
                    translate("Arch", "The chosen object is not a Structure") + "\n"
                )
            else:
                if not other.Nodes:
                    FreeCAD.Console.PrintError(
                        translate("Arch", "The chosen object has no structural nodes") + "\n"
                    )
                else:
                    if (len(self.Object.Nodes) != 2) or (len(other.Nodes) != 2):
                        FreeCAD.Console.PrintError(
                            translate("Arch", "One of these objects has more than 2 nodes") + "\n"
                        )
                    else:
                        import DraftGeomUtils

                        nodes1 = [self.Object.Placement.multVec(v) for v in self.Object.Nodes]
                        nodes2 = [other.Placement.multVec(v) for v in other.Nodes]
                        intersect = DraftGeomUtils.findIntersection(
                            nodes1[0], nodes1[1], nodes2[0], nodes2[1], True, True
                        )
                        if not intersect:
                            FreeCAD.Console.PrintError(
                                translate("Arch", "Unable to find a suitable intersection point")
                                + "\n"
                            )
                        else:
                            intersect = intersect[0]
                            FreeCAD.Console.PrintMessage(translate("Arch", "Intersection found.\n"))
                            if DraftGeomUtils.findClosest(intersect, nodes1) == 0:
                                self.Object.Nodes = [
                                    self.Object.Placement.inverse().multVec(intersect),
                                    self.Object.Nodes[1],
                                ]
                            else:
                                self.Object.Nodes = [
                                    self.Object.Nodes[0],
                                    self.Object.Placement.inverse().multVec(intersect),
                                ]

    def connectNodes(self, other=None):

        if not other:
            self.observer = StructSelectionObserver(self.connectNodes)
            FreeCADGui.Selection.addObserver(self.observer)
            FreeCAD.Console.PrintMessage(translate("Arch", "Choose another Structure object:"))
        else:
            FreeCADGui.Selection.removeObserver(self.observer)
            self.observer = None
            if Draft.getType(other) != "Structure":
                FreeCAD.Console.PrintError(
                    translate("Arch", "The chosen object is not a Structure") + "\n"
                )
            else:
                if not other.Nodes:
                    FreeCAD.Console.PrintError(
                        translate("Arch", "The chosen object has no structural nodes") + "\n"
                    )
                else:
                    if (len(self.Object.Nodes) != 2) or (len(other.Nodes) != 2):
                        FreeCAD.Console.PrintError(
                            translate("Arch", "One of these objects has more than 2 nodes") + "\n"
                        )
                    else:
                        import DraftGeomUtils

                        nodes1 = [self.Object.Placement.multVec(v) for v in self.Object.Nodes]
                        nodes2 = [other.Placement.multVec(v) for v in other.Nodes]
                        intersect = DraftGeomUtils.findIntersection(
                            nodes1[0], nodes1[1], nodes2[0], nodes2[1], True, True
                        )
                        if not intersect:
                            FreeCAD.Console.PrintError(
                                translate("Arch", "Unable to find a suitable intersection point")
                                + "\n"
                            )
                        else:
                            intersect = intersect[0]
                            FreeCAD.Console.PrintMessage(
                                translate("Arch", "Intersection found.") + "\n"
                            )
                            if DraftGeomUtils.findClosest(intersect, nodes1) == 0:
                                self.Object.Nodes = [
                                    self.Object.Placement.inverse().multVec(intersect),
                                    self.Object.Nodes[1],
                                ]
                            else:
                                self.Object.Nodes = [
                                    self.Object.Nodes[0],
                                    self.Object.Placement.inverse().multVec(intersect),
                                ]
                            if DraftGeomUtils.findClosest(intersect, nodes2) == 0:
                                other.Nodes = [
                                    other.Placement.inverse().multVec(intersect),
                                    other.Nodes[1],
                                ]
                            else:
                                other.Nodes = [
                                    other.Nodes[0],
                                    other.Placement.inverse().multVec(intersect),
                                ]

    def toggleNodes(self):

        if self.nodevis:
            for obj in self.nodevis:
                obj[0].ViewObject.ShowNodes = obj[1]
            self.nodevis = None
        else:
            self.nodevis = []
            for obj in FreeCAD.ActiveDocument.Objects:
                if hasattr(obj.ViewObject, "ShowNodes"):
                    self.nodevis.append([obj, obj.ViewObject.ShowNodes])
                    obj.ViewObject.ShowNodes = True

    def setSelectionFromTool(self):
        FreeCADGui.Selection.clearSelection()
        if hasattr(self.Object, "Tool"):
            tool = self.Object.Tool
            if hasattr(tool, "Shape") and tool.Shape:
                FreeCADGui.Selection.addSelection(tool)
            else:
                if not isinstance(tool, list):
                    tool = [tool]
                for o, subs in tool:
                    FreeCADGui.Selection.addSelection(o, subs)
        QtCore.QObject.disconnect(
            self.selectToolButton, QtCore.SIGNAL("clicked()"), self.setSelectionFromTool
        )
        QtCore.QObject.connect(
            self.selectToolButton, QtCore.SIGNAL("clicked()"), self.setToolFromSelection
        )
        self.selectToolButton.setText(QtGui.QApplication.translate("Arch", "Done", None))

    def setToolFromSelection(self):
        objectList = []
        selEx = FreeCADGui.Selection.getSelectionEx()
        for selExi in selEx:
            if len(selExi.SubElementNames) == 0:
                # Add entirely selected objects
                objectList.append(selExi.Object)
            else:
                subElementsNames = [
                    subElementName
                    for subElementName in selExi.SubElementNames
                    if subElementName.startswith("Edge")
                ]
                # Check that at least an edge is selected from the object's shape
                if len(subElementsNames) > 0:
                    objectList.append((selExi.Object, subElementsNames))
        if self.Object.getTypeIdOfProperty("Tool") != "App::PropertyLinkSubList":
            # Upgrade property Tool from App::PropertyLink to App::PropertyLinkSubList (note: Undo/Redo fails)
            self.Object.removeProperty("Tool")
            self.Object.addProperty(
                "App::PropertyLinkSubList",
                "Tool",
                "Structure",
                QT_TRANSLATE_NOOP("App::Property", "An optional extrusion path for this element"),
                locked=True,
            )
        self.Object.Tool = objectList
        QtCore.QObject.disconnect(
            self.selectToolButton, QtCore.SIGNAL("clicked()"), self.setToolFromSelection
        )
        QtCore.QObject.connect(
            self.selectToolButton, QtCore.SIGNAL("clicked()"), self.setSelectionFromTool
        )
        self.selectToolButton.setText(QtGui.QApplication.translate("Arch", "Select Tool", None))

    def accept(self):

        if self.observer:
            FreeCADGui.Selection.removeObserver(self.observer)
        if self.nodevis:
            self.toggleNodes()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True


class StructSelectionObserver:

    def __init__(self, callback):
        self.callback = callback

    def addSelection(self, docName, objName, sub, pos):
        print("got ", objName)
        obj = FreeCAD.getDocument(docName).getObject(objName)
        self.callback(obj)


class _StructuralSystem(
    ArchComponent.Component
):  # OBSOLETE - All Arch objects can now be based on axes
    "The Structural System object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "StructuralSystem"

        obj.addProperty(
            "App::PropertyLinkList",
            "Axes",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "Axes systems this structure is built on"),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyIntegerList",
            "Exclude",
            "Arch",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The element numbers to exclude when this structure is based on axes",
            ),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyBool",
            "Align",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "If true the element are aligned with axes"),
            locked=True,
        ).Align = False

    def loads(self, state):

        self.Type = "StructuralSystem"

    def execute(self, obj):
        "creates the structure shape"

        import Part
        import DraftGeomUtils

        # creating base shape
        pl = obj.Placement
        if obj.Base:
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape.isNull():
                    return
                if not obj.Base.Shape.Solids:
                    return

                base = None

                # applying axes
                pts = self.getAxisPoints(obj)
                if hasattr(obj, "Align"):
                    if obj.Align == False:
                        apl = self.getAxisPlacement(obj)
                    if obj.Align:
                        apl = None
                else:
                    apl = self.getAxisPlacement(obj)

                if pts:
                    fsh = []
                    for i in range(len(pts)):
                        sh = obj.Base.Shape.copy()
                        if hasattr(obj, "Exclude"):
                            if i in obj.Exclude:
                                continue
                        if apl:
                            sh.Placement.Rotation = sh.Placement.Rotation.multiply(apl.Rotation)
                        sh.translate(pts[i])
                        fsh.append(sh)

                    if fsh:
                        base = Part.makeCompound(fsh)
                        base = self.processSubShapes(obj, base, pl)

                if base:
                    if not base.isNull():
                        if base.isValid() and base.Solids:
                            if base.Volume < 0:
                                base.reverse()
                            if base.Volume < 0:
                                FreeCAD.Console.PrintError(
                                    translate("Arch", "Could not compute a shape")
                                )
                                return
                            base = base.removeSplitter()
                            obj.Shape = base
                            if not pl.isNull():
                                obj.Placement = pl

    def getAxisPoints(self, obj):
        "returns the gridpoints of linked axes"
        import DraftGeomUtils

        pts = []
        if len(obj.Axes) == 1:
            if hasattr(obj, "Align"):
                if obj.Align:
                    p0 = obj.Axes[0].Shape.Edges[0].Vertexes[1].Point
                    for e in obj.Axes[0].Shape.Edges:
                        p = e.Vertexes[1].Point
                        p = p.sub(p0)
                        pts.append(p)
                else:
                    for e in obj.Axes[0].Shape.Edges:
                        pts.append(e.Vertexes[0].Point)
            else:
                for e in obj.Axes[0].Shape.Edges:
                    pts.append(e.Vertexes[0].Point)
        elif len(obj.Axes) >= 2:
            set1 = obj.Axes[0].Shape.Edges
            set2 = obj.Axes[1].Shape.Edges
            for e1 in set1:
                for e2 in set2:
                    pts.extend(DraftGeomUtils.findIntersection(e1, e2))
        return pts

    def getAxisPlacement(self, obj):
        "returns an axis placement"
        if obj.Axes:
            return obj.Axes[0].Placement
        return None


class _ViewProviderStructuralSystem(ArchComponent.ViewProviderComponent):
    "A View Provider for the Structural System object"

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_StructuralSystem_Tree.svg"


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("Arch_Structure", _CommandStructure())
    FreeCADGui.addCommand("Arch_StructuralSystem", CommandStructuralSystem())
    FreeCADGui.addCommand("Arch_StructuresFromSelection", CommandStructuresFromSelection())

    class _ArchStructureGroupCommand:

        def GetCommands(self):
            return ("Arch_Structure", "Arch_StructuralSystem", "Arch_StructuresFromSelection")

        def GetResources(self):
            return {
                "MenuText": QT_TRANSLATE_NOOP("Arch_StructureTools", "Structure Tools"),
                "ToolTip": QT_TRANSLATE_NOOP("Arch_StructureTools", "Structure tools"),
            }

        def IsActive(self):
            return not FreeCAD.ActiveDocument is None

    FreeCADGui.addCommand("Arch_StructureTools", _ArchStructureGroupCommand())



# /******************************************************************************
# *   Copyright (c) 2012 Jan Rheinländer <jrheinlaender@users.sourceforge.net> *
# *                                                                            *
# *   This file is part of the FreeCAD CAx development system.                 *
# *                                                                            *
# *   This library is free software; you can redistribute it and/or            *
# *   modify it under the terms of the GNU Library General Public              *
# *   License as published by the Free Software Foundation; either             *
# *   version 2 of the License, or (at your option) any later version.         *
# *                                                                            *
# *   This library  is distributed in the hope that it will be useful,         *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
# *   GNU Library General Public License for more details.                     *
# *                                                                            *
# *   You should have received a copy of the GNU Library General Public        *
# *   License along with this library; see the file COPYING.LIB. If not,       *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,            *
# *   Suite 330, Boston, MA  02111-1307, USA                                   *
# *                                                                            *
# ******************************************************************************/

import FreeCAD, FreeCADGui
import traceback
from PySide import QtCore, QtGui
from .WizardShaftTable import WizardShaftTable
from .Shaft import Shaft

translate = FreeCAD.Qt.translate


class TaskWizardShaft:
    "Shaft Wizard"

    App = FreeCAD
    Gui = FreeCADGui

    def __init__(self, doc):
        mw = QtGui.QApplication.activeWindow()
        # cw = mw.centralWidget() # This is a qmdiarea widget
        cw = mw.findChild(QtGui.QMdiArea)
        self.doc = doc

        # Get active document or create a new one
        # Important because when setting the default data in WizardShaftTable() the
        # updateSketch() slot will be activated and it relies on finding a valid document
        if self.doc is None:
            self.Gui.activateWorkbench("PartDesignWorkbench")
            self.doc = self.App.newDocument()
            # Grab the newly created feature window
            featureWindow = cw.subWindowList()[-1]
        else:
            featureWindow = cw.activeSubWindow()

        # Buttons for diagram display
        buttonLayout = QtGui.QGridLayout()
        all = translate("TaskWizardShaft", "All")
        bnames = [
            [f"{all} [x]", f"{all} [y]", f"{all} [z]"],
            ["N [x]", "Q [y]", "Q [z]"],
            ["Mt [x]", "Mb [z]", "Mb [y]"],
            ["", "w [y]", "w [z]"],
            ["sigma [x]", "sigma [y]", "sigma [z]"],
            ["tau [x]", "sigmab [z]", "sigmab [y]"],
        ]
        slots = [
            [self.slotAllx, self.slotAlly, self.slotAllz],
            [self.slotFx, self.slotQy, self.slotQz],
            [self.slotMx, self.slotMz, self.slotMy],
            [self.slotNone, self.slotWy, self.slotWz],
            [self.slotSigmax, self.slotSigmay, self.slotSigmaz],
            [self.slotTaut, self.slotSigmabz, self.slotSigmaby],
        ]
        self.buttons = [
            [None, None, None],
            [None, None, None],
            [None, None, None],
            [None, None, None],
            [None, None, None],
            [None, None, None],
        ]

        for col in range(3):
            for row in range(6):
                button = QtGui.QPushButton(bnames[row][col])
                buttonLayout.addWidget(button, row, col)
                self.buttons[row][col] = button
                button.clicked.connect(slots[row][col])

        # Create Shaft object
        self.shaft = Shaft(self)
        # Create table widget
        self.form = QtGui.QWidget()
        self.table = WizardShaftTable(self, self.shaft)

        # The top layout will contain the Shaft Wizard layout plus the elements of the FEM constraints dialog
        layout = QtGui.QVBoxLayout()
        layout.setObjectName(
            "ShaftWizard"
        )  # Do not change or translate: Required to detect whether Shaft Wizard is running in FemGui::ViewProviderFemConstraintXXX
        sublayout = QtGui.QVBoxLayout()
        sublayout.setObjectName("ShaftWizardLayout")  # Do not change or translate
        sublayout.addWidget(self.table.widget)
        sublayout.addLayout(buttonLayout)
        layout.addLayout(sublayout)
        self.form.setLayout(layout)

        # Switch to feature window
        mdi = FreeCADGui.getMainWindow().findChild(QtGui.QMdiArea)
        cw.setActiveSubWindow(featureWindow)

    def showDiagram(self, diagram):
        try:
            self.shaft.showDiagram(diagram)
        except ImportError as e:
            msgBox = QtGui.QMessageBox()
            msgBox.setIcon(msgBox.Information)
            msgBox.setWindowTitle(translate("TaskWizardShaft", "Missing Module"))
            msgBox.setText(
                translate(
                    "TaskWizardShaft",
                    "The Plot add-on is not installed. Install it to enable this feature.",
                )
            )
            msgBox.setDetailedText(traceback.format_exc())
            msgBox.exec_()

    def slotAllx(self):
        self.showDiagram("Allx")

    def slotAlly(self):
        self.showDiagram("Ally")

    def slotAllz(self):
        self.showDiagram("Allz")

    def slotFx(self):
        self.showDiagram("Nx")

    def slotQy(self):
        self.showDiagram("Qy")

    def slotQz(self):
        self.showDiagram("Qz")

    def slotMx(self):
        self.showDiagram("Mx")

    def slotMz(self):
        self.showDiagram("Mz")

    def slotMy(self):
        self.showDiagram("My")

    def slotNone(self):
        pass

    def slotWy(self):
        self.showDiagram("wy")

    def slotWz(self):
        self.showDiagram("wz")

    def slotSigmax(self):
        self.showDiagram("sigmax")

    def slotSigmay(self):
        self.showDiagram("sigmay")

    def slotSigmaz(self):
        self.showDiagram("sigmaz")

    def slotTaut(self):
        self.showDiagram("taut")

    def slotSigmabz(self):
        self.showDiagram("sigmabz")

    def slotSigmaby(self):
        self.showDiagram("sigmaby")

    def updateButton(self, row, col, flag):
        self.buttons[row][col].setEnabled(flag)

    def updateButtons(self, col, flag):
        for row in range(len(self.buttons)):
            self.updateButton(row, col, flag)

    def getStandardButtons(self):
        return QtGui.QDialogButtonBox.Ok

    def accept(self):
        if self.table:
            del self.table
        if self.shaft:
            del self.shaft
        if self.form:
            del self.form
        return True

    def isAllowedAlterDocument(self):
        return False


# HACK: Workaround to allow a callback
# Problem: From the FemConstraint ViewProvider, we need to tell the Shaft instance that the user finished editing the constraint
# We can find the Shaft Wizard dialog object from C++, but there is no way to reach the Shaft instance
# Also it seems to be impossible to access the active dialog from Python, so Gui::Command::runCommand() is not an option either
# Note: Another way would be to create a hidden widget in the Shaft Wizard dialog and write some data to it, triggering a slot
# in the python code
WizardShaftDlg = None


class WizardShaftGui:
    def Activated(self):
        global WizardShaftDlg
        WizardShaftDlg = TaskWizardShaft(FreeCAD.ActiveDocument)
        FreeCADGui.Control.showDialog(WizardShaftDlg)

    def GetResources(self):
        IconPath = FreeCAD.ConfigGet("AppHomePath") + "Mod/PartDesign/WizardShaft/WizardShaft.svg"
        MenuText = QtCore.QT_TRANSLATE_NOOP("PartDesign_WizardShaft", "Shaft Design Wizard")
        ToolTip = QtCore.QT_TRANSLATE_NOOP(
            "PartDesign_WizardShaft", "Starts the shaft design wizard"
        )
        return {"Pixmap": IconPath, "MenuText": MenuText, "ToolTip": ToolTip}

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def __del__(self):
        global WizardShaftDlg
        WizardShaftDlg = None


class WizardShaftGuiCallback:
    def Activated(self):
        global WizardShaftDlg
        if WizardShaftDlg is not None and WizardShaftDlg.table is not None:
            WizardShaftDlg.table.finishEditConstraint()

    def isActive(self):
        global WizardShaftDlg
        return WizardShaftDlg is not None

    def GetResources(self):
        IconPath = FreeCAD.ConfigGet("AppHomePath") + "Mod/PartDesign/WizardShaft/WizardShaft.svg"
        MenuText = QtCore.QT_TRANSLATE_NOOP(
            "PartDesign_WizardShaftCallBack", "Shaft design wizard..."
        )
        ToolTip = QtCore.QT_TRANSLATE_NOOP(
            "PartDesign_WizardShaftCallBack", "Start the shaft design wizard"
        )
        return {"Pixmap": IconPath, "MenuText": MenuText, "ToolTip": ToolTip}


FreeCADGui.addCommand("PartDesign_WizardShaft", WizardShaftGui())
FreeCADGui.addCommand("PartDesign_WizardShaftCallBack", WizardShaftGuiCallback())

# Note: Start wizard in Python Console with
# Gui.runCommand('PartDesign_WizardShaft')



#!/usr/bin/env python3
# SPDX-License-Identifier: LGPL-2.1-or-later

#  Copyright (c) 2007 Jürgen Riegel <juergen.riegel@web.de>
#  LGPL

import os
import sys
import io
import FreeCAD, unittest, Mesh
import MeshEnums
from FreeCAD import Base
import time, tempfile, math

# http://python-kurs.eu/threads.php
try:
    import _thread as thread
except Exception:
    import thread

from os.path import join

# ---------------------------------------------------------------------------
# define the functions to test the FreeCAD mesh module
# ---------------------------------------------------------------------------


class MeshTopoTestCases(unittest.TestCase):
    def setUp(self):
        # set up a planar face with 18 triangles
        self.planarMesh = []
        for x in range(3):
            for y in range(3):
                self.planarMesh.append([0.0 + x, 0.0 + y, 0.0000])
                self.planarMesh.append([1.0 + x, 1.0 + y, 0.0000])
                self.planarMesh.append([0.0 + x, 1.0 + y, 0.0000])
                self.planarMesh.append([0.0 + x, 0.0 + y, 0.0000])
                self.planarMesh.append([1.0 + x, 0.0 + y, 0.0000])
                self.planarMesh.append([1.0 + x, 1.0 + y, 0.0000])

    def testCollapseFacetsSingle(self):
        for i in range(18):
            planarMeshObject = Mesh.Mesh(self.planarMesh)
            planarMeshObject.collapseFacets([i])

    def testCollapseFacetsMultible(self):
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        planarMeshObject.collapseFacets(range(7))

    def testCollapseFacetsAll(self):
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        planarMeshObject.collapseFacets(range(18))

    # fmt: off
    def testCorruptedFacet(self):
        v = FreeCAD.Vector
        mesh = Mesh.Mesh()
        mesh.addFacet(
        v(1.0e1, -1.0e1, 1.0e1),
        v(1.0e1, +1.0e1, 1.0e1),
        v(0.0e0, 0.0e0, 1.0e1))

        mesh.addFacet(
        v(-1.0e1, -1.0e1, 1.0e1),
        v(-1.0e1, +1.0e1, 1.0e1),
        v(0e0, 0.0e0, 1.0e1))

        mesh.addFacet(
        v(+1.0e1, +1.0e1, 1.0e1),
        v(-1.0e1, +1.0e1, 1.0e1),
        v(.0e0, 0.0e0, 1.0e1))

        mesh.addFacet(
        v(+1.0e1, -1.0e1, 1.0e1),
        v(-1.0e1, -1.0e1, 1.0e1),
        v(.0e0, 0.0e0, 1.0e1))

        mesh.addFacet(
        v(-1.0e1, +1.0e1, 1.0e1),
        v(+1.0e1, +1.0e1, 1.0e1),
        v(+1.0e1, +1.0e1, 1.0e1))

        mesh.addFacet(
        v(+1.0e1, +1.0e1, 1.0e1),
        v(+1.0e1, 00.0e1, 1.0e1),
        v(+1.0e1, -1.0e1, 1.0e1))

        self.assertEqual(mesh.CountFacets, 6)
        mesh.fixIndices()
        self.assertEqual(mesh.CountFacets, 5)


# fmt: on


class MeshSplitTestCases(unittest.TestCase):
    def setUp(self):
        self.mesh = Mesh.createBox(1.0, 1.0, 1.0)

    def testSplitFacetOnOneEdge(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, p1, (p2 + p3) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_21(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p1 + p3) / 2, (p2 + p3) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_12(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p2 + p3) / 2, (p1 + p3) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_01(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p1 + p2) / 2, (p2 + p3) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_10(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p2 + p3) / 2, (p1 + p2) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_02(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p1 + p2) / 2, (p1 + p3) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_20(self):
        p1 = self.mesh.Points[0].Vector
        p2 = self.mesh.Points[1].Vector
        p3 = self.mesh.Points[2].Vector
        self.mesh.splitFacet(0, (p1 + p3) / 2, (p1 + p2) / 2)
        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testSplitFacetOnTwoEdges_5teps(self):
        Vec3d = FreeCAD.Vector
        for i in range(5):
            f = self.mesh.Facets[0]
            p1 = Vec3d(f.Points[0])
            p2 = Vec3d(f.Points[1])
            p3 = Vec3d(f.Points[2])
            self.mesh.splitFacet(0, (p1 + p3) / 2, (p2 + p3) / 2)

        self.assertFalse(self.mesh.hasNonManifolds())
        self.assertFalse(self.mesh.hasInvalidNeighbourhood())
        self.assertFalse(self.mesh.hasPointsOutOfRange())
        self.assertFalse(self.mesh.hasFacetsOutOfRange())
        self.assertFalse(self.mesh.hasCorruptedFacets())
        self.assertTrue(self.mesh.isSolid())

    def testFindNearest(self):
        self.assertEqual(len(self.mesh.nearestFacetOnRay((-2, 2, -6), (0, 0, 1))), 0)
        self.assertEqual(len(self.mesh.nearestFacetOnRay((0.5, 0.5, 0.5), (0, 0, 1))), 1)
        self.assertEqual(
            len(self.mesh.nearestFacetOnRay((0.5, 0.5, 0.5), (0, 0, 1), -math.pi / 2)), 0
        )
        self.assertEqual(
            len(self.mesh.nearestFacetOnRay((0.2, 0.1, 0.2), (0, 0, 1))),
            len(self.mesh.nearestFacetOnRay((0.2, 0.1, 0.2), (0, 0, -1))),
        )
        self.assertEqual(
            len(self.mesh.nearestFacetOnRay((0.2, 0.1, 0.2), (0, 0, 1), math.pi / 2)),
            len(self.mesh.nearestFacetOnRay((0.2, 0.1, 0.2), (0, 0, -1), math.pi / 2)),
        )
        # Apply placement to mesh
        plm = Base.Placement(Base.Vector(1, 2, 3), Base.Rotation(1, 1, 1, 1))
        pnt = Base.Vector(0.5, 0.5, 0.5)
        vec = Base.Vector(0.0, 0.0, 1.0)

        self.mesh.Placement = plm
        self.assertEqual(len(self.mesh.nearestFacetOnRay(pnt, vec)), 0)

        # Apply the placement on the ray as well
        pnt = plm.multVec(pnt)
        vec = plm.Rotation.multVec(vec)
        self.assertEqual(len(self.mesh.nearestFacetOnRay(pnt, vec)), 1)

    def testForaminate(self):
        class FilterAngle:
            def __init__(self, mesh, vec, limit):
                self.myMesh = mesh
                self.vec = vec
                self.limit = limit

            def check_angle(self, item):
                angle = self.myMesh.Facets[item].Normal.getAngle(self.vec)
                return angle < self.limit

        results = self.mesh.foraminate((0.0, 0.0, 0.0), (0, 1, 1))
        filtered_result = list(
            filter(
                FilterAngle(self.mesh, FreeCAD.Vector(0, 1, 1), math.pi / 2).check_angle,
                results.keys(),
            )
        )

        self.assertEqual(
            filtered_result, list(self.mesh.foraminate((0.0, 0.0, 0.0), (0, 1, 1), math.pi / 2))
        )

    def testForaminatePlacement(self):
        pnt = Base.Vector(0.0, 0.0, 0.0)
        vec = Base.Vector(0.0, 1.0, 1.0)
        results = self.mesh.foraminate(pnt, vec)
        self.assertEqual(len(results), 4)

        # Apply placement to mesh
        plm = Base.Placement(Base.Vector(1, 2, 3), Base.Rotation(1, 1, 1, 1))
        self.mesh.Placement = plm
        self.assertEqual(len(self.mesh.foraminate(pnt, vec)), 0)

        # Apply the placement on the ray as well
        pnt = plm.multVec(pnt)
        vec = plm.Rotation.multVec(vec)
        results2 = self.mesh.foraminate(pnt, vec)
        self.assertEqual(len(results2), 4)
        self.assertEqual(list(results), list(results2))


class MeshGeoTestCases(unittest.TestCase):
    def setUp(self):
        # set up a planar face with 2 triangles
        self.planarMesh = []

    def testIntersection(self):
        self.planarMesh.append([0.9961, 1.5413, 4.3943])
        self.planarMesh.append([9.4796, 10.024, -3.0937])
        self.planarMesh.append([1.4308, 11.3841, 2.6829])
        self.planarMesh.append([2.6493, 2.2536, 3.0679])
        self.planarMesh.append([13.1126, 0.4857, -4.4417])
        self.planarMesh.append([10.2410, 8.9040, -3.5002])
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertTrue(len(res) == 0)

    def testIntersection2(self):
        self.planarMesh.append([-16.097176, -29.891157, 15.987688])
        self.planarMesh.append([-16.176304, -29.859991, 15.947966])
        self.planarMesh.append([-16.071451, -29.900553, 15.912505])
        self.planarMesh.append([-16.092241, -29.893408, 16.020439])
        self.planarMesh.append([-16.007210, -29.926180, 15.967641])
        self.planarMesh.append([-16.064457, -29.904951, 16.090832])
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        # does definitely NOT intersect
        res = f1.intersect(f2)
        self.assertTrue(len(res) == 0)

    def testIntersectionOfTransformedMesh(self):
        self.planarMesh.append([0.0, 10.0, 10.0])
        self.planarMesh.append([10.0, 0.0, 10.0])
        self.planarMesh.append([10.0, 10.0, 10.0])
        self.planarMesh.append([6.0, 8.0, 10.0])
        self.planarMesh.append([16.0, 8.0, 10.0])
        self.planarMesh.append([6.0, 18.0, 10.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        mat = Base.Matrix()
        mat.rotateX(1.0)
        mat.rotateY(1.0)
        mat.rotateZ(1.0)
        planarMeshObject.transformGeometry(mat)

        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertEqual(len(res), 2)

    def testIntersectionOfParallelTriangles(self):
        self.planarMesh.append([0.0, 10.0, 10.0])
        self.planarMesh.append([10.0, 0.0, 10.0])
        self.planarMesh.append([10.0, 10.0, 10.0])
        self.planarMesh.append([6.0, 8.0, 10.1])
        self.planarMesh.append([16.0, 8.0, 10.1])
        self.planarMesh.append([6.0, 18.0, 10.1])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        mat = Base.Matrix()
        mat.rotateX(1.0)
        mat.rotateY(1.0)
        mat.rotateZ(1.0)
        planarMeshObject.transformGeometry(mat)

        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertTrue(len(res) == 0)

    def testIntersectionOnEdge(self):
        self.planarMesh.append([5.0, -1.9371663331985474, 0.49737977981567383])
        self.planarMesh.append([4.0, -1.9371663331985474, 0.49737977981567383])
        self.planarMesh.append([5.0, -1.9842294454574585, 0.25066646933555603])
        self.planarMesh.append([4.6488823890686035, -1.7827962636947632, 0.4577442705631256])
        self.planarMesh.append([4.524135112762451, -2.0620131492614746, 0.5294350385665894])
        self.planarMesh.append([4.6488823890686035, -1.8261089324951172, 0.23069120943546295])
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertEqual(len(res), 2)

    def testIntersectionCoplanar(self):
        self.planarMesh.append([0.0, 10.0, 10.0])
        self.planarMesh.append([10.0, 0.0, 10.0])
        self.planarMesh.append([10.0, 10.0, 10.0])
        self.planarMesh.append([6.0, 8.0, 10.0])
        self.planarMesh.append([16.0, 8.0, 10.0])
        self.planarMesh.append([6.0, 18.0, 10.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertTrue(len(res) == 2)

    def testIntersectionOverlap(self):
        self.planarMesh.append([0.0, 0.0, 0.0])
        self.planarMesh.append([5.0, 0.0, 0.0])
        self.planarMesh.append([8.0, 5.0, 0.0])
        self.planarMesh.append([4.0, 0.0, 0.0])
        self.planarMesh.append([10.0, 0.0, 0.0])
        self.planarMesh.append([9.0, 5.0, 0.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)
        f1 = planarMeshObject.Facets[0]
        f2 = planarMeshObject.Facets[1]
        res = f1.intersect(f2)
        self.assertTrue(len(res) == 2)

    def testIntersectionOfIntersectingEdges(self):
        self.planarMesh.append([0.0, 10.0, 10.0])
        self.planarMesh.append([10.0, 0.0, 10.0])
        self.planarMesh.append([10.0, 10.0, 10.0])
        self.planarMesh.append([6.0, 8.0, 10.0])
        self.planarMesh.append([16.0, 8.0, 10.0])
        self.planarMesh.append([6.0, 18.0, 10.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        edge1 = planarMeshObject.Facets[0].getEdge(2)
        edge2 = planarMeshObject.Facets[1].getEdge(2)
        res = edge1.intersectWithEdge(edge2)
        self.assertTrue(len(res) == 1)
        self.assertEqual(res[0][0], 6.0)
        self.assertEqual(res[0][1], 10.0)
        self.assertEqual(res[0][2], 10.0)

    def testIntersectionOfParallelEdges(self):
        self.planarMesh.append([0.0, 10.0, 10.0])
        self.planarMesh.append([10.0, 0.0, 10.0])
        self.planarMesh.append([10.0, 10.0, 10.0])
        self.planarMesh.append([6.0, 8.0, 10.0])
        self.planarMesh.append([16.0, 8.0, 10.0])
        self.planarMesh.append([6.0, 18.0, 10.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        edge1 = planarMeshObject.Facets[0].getEdge(2)
        edge2 = planarMeshObject.Facets[1].getEdge(0)
        res = edge1.intersectWithEdge(edge2)
        self.assertTrue(len(res) == 0)

    def testIntersectionOfCollinearEdges(self):
        self.planarMesh.append([0.0, 0.0, 0.0])
        self.planarMesh.append([6.0, 0.0, 0.0])
        self.planarMesh.append([3.0, 4.0, 0.0])
        self.planarMesh.append([7.0, 0.0, 0.0])
        self.planarMesh.append([13.0, 0.0, 0.0])
        self.planarMesh.append([10.0, 4.0, 0.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        edge1 = planarMeshObject.Facets[0].getEdge(0)
        edge2 = planarMeshObject.Facets[1].getEdge(0)
        res = edge1.intersectWithEdge(edge2)
        self.assertTrue(len(res) == 0)

    def testIntersectionOfWarpedEdges(self):
        self.planarMesh.append([0.0, 0.0, 0.0])
        self.planarMesh.append([6.0, 0.0, 0.0])
        self.planarMesh.append([3.0, 4.0, 0.0])
        self.planarMesh.append([2.0, 2.0, 1.0])
        self.planarMesh.append([8.0, 2.0, 1.0])
        self.planarMesh.append([5.0, 6.0, 1.0])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        edge1 = planarMeshObject.Facets[0].getEdge(1)
        edge2 = planarMeshObject.Facets[1].getEdge(0)
        res = edge1.intersectWithEdge(edge2)
        self.assertTrue(len(res) == 0)

    def testSelfIntersection(self):
        s = b"""solid Simple
facet normal 0.0e0 0.0e0 1.0e1
    outer loop
        vertex 0.0e1 0.0e1 1.0e1
        vertex 0.0e1 +1.0e1 1.0e1
        vertex +1.0e1 0.0e1 1.0e1
    endloop
endfacet
facet normal 0.0e0 0.0e0 1.0e1
    outer loop
        vertex 0.0e1 +1.0e1 1.0e1
        vertex +1.0e1 0.0e1 1.0e1
        vertex 1.0e1 1.0e1 1.0e1
    endloop
endfacet
facet normal 0.0e0 0.0e0 1.0e1
    outer loop
        vertex 0.0e1 0.0e1 1.0e1
        vertex 0.0e1 +1.0e1 1.0e1
        vertex -1.0e1 1.0e1 1.0e1
    endloop
endfacet
facet normal 0.0e0 0.0e0 1.0e1
    outer loop
        vertex 0.0e1 0.0e1 1.0e1
        vertex +1.0e1 0.0e1 1.0e1
        vertex +1.0e1 -1.0e1 1.0e1
    endloop
endfacet
facet normal 0.0e0 0.0e0 1.0e1
    outer loop
        vertex 0.6e1 0.8e1 1.0e1
        vertex +1.6e1 0.8e1 1.0e1
        vertex +0.6e1 1.8e1 1.0e1
    endloop
endfacet
endsolid Simple"""
        mesh = Mesh.Mesh()
        data = io.BytesIO(s)
        mesh.read(Stream=data, Format="AST")
        self.assertTrue(mesh.hasSelfIntersections())


class PivyTestCases(unittest.TestCase):
    def setUp(self):
        # set up a planar face with 2 triangles
        self.planarMesh = []
        FreeCAD.newDocument("MeshTest")

    def testRayPick(self):
        if not FreeCAD.GuiUp:
            return
        self.planarMesh.append([-16.097176, -29.891157, 15.987688])
        self.planarMesh.append([-16.176304, -29.859991, 15.947966])
        self.planarMesh.append([-16.071451, -29.900553, 15.912505])
        self.planarMesh.append([-16.092241, -29.893408, 16.020439])
        self.planarMesh.append([-16.007210, -29.926180, 15.967641])
        self.planarMesh.append([-16.064457, -29.904951, 16.090832])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        from pivy import coin
        import FreeCADGui

        Mesh.show(planarMeshObject)
        view = FreeCADGui.ActiveDocument.ActiveView.getViewer()
        rp = coin.SoRayPickAction(view.getSoRenderManager().getViewportRegion())
        rp.setRay(coin.SbVec3f(-16.05, 16.0, 16.0), coin.SbVec3f(0, -1, 0))
        rp.apply(view.getSoRenderManager().getSceneGraph())
        pp = rp.getPickedPoint()
        self.assertTrue(pp is not None)
        det = pp.getDetail()
        self.assertTrue(det.getTypeId() == coin.SoFaceDetail.getClassTypeId())
        det = coin.cast(det, det.getTypeId().getName().getString())
        self.assertTrue(det.getFaceIndex() == 1)

    def testPrimitiveCount(self):
        if not FreeCAD.GuiUp:
            return
        self.planarMesh.append([-16.097176, -29.891157, 15.987688])
        self.planarMesh.append([-16.176304, -29.859991, 15.947966])
        self.planarMesh.append([-16.071451, -29.900553, 15.912505])
        self.planarMesh.append([-16.092241, -29.893408, 16.020439])
        self.planarMesh.append([-16.007210, -29.926180, 15.967641])
        self.planarMesh.append([-16.064457, -29.904951, 16.090832])
        planarMeshObject = Mesh.Mesh(self.planarMesh)

        from pivy import coin
        import FreeCADGui

        Mesh.show(planarMeshObject)
        view = FreeCADGui.ActiveDocument.ActiveView
        view.setAxisCross(False)
        pc = coin.SoGetPrimitiveCountAction()
        pc.apply(view.getSceneGraph())
        self.assertTrue(pc.getTriangleCount() == 2)
        # self.assertTrue(pc.getPointCount() == 6)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("MeshTest")


# Threads


def loadFile(name):
    # lock.acquire()
    mesh = Mesh.Mesh()
    # FreeCAD.Console.PrintMessage("Create mesh instance\n")
    # lock.release()
    mesh.read(name)
    # FreeCAD.Console.PrintMessage("Mesh loaded successfully.\n")


def createMesh(r, s):
    # FreeCAD.Console.PrintMessage("Create sphere (%s,%s)...\n"%(r,s))
    mesh = Mesh.createSphere(r, s)
    # FreeCAD.Console.PrintMessage("... destroy sphere\n")


class LoadMeshInThreadsCases(unittest.TestCase):
    def setUp(self):
        pass

    def testSphereMesh(self):
        for i in range(6, 8):
            thread.start_new(createMesh, (10.0, (i + 1) * 20))
        time.sleep(10)

    def testLoadMesh(self):
        mesh = Mesh.createSphere(10.0, 100)  # a fine sphere
        name = tempfile.gettempdir() + os.sep + "mesh.stl"
        mesh.write(name)
        # FreeCAD.Console.PrintMessage("Write mesh to %s\n"%(name))
        # lock=thread.allocate_lock()
        for i in range(2):
            thread.start_new(loadFile, (name,))
        time.sleep(1)

    def tearDown(self):
        pass


class PolynomialFitCases(unittest.TestCase):
    def setUp(self):
        pass

    def testFitGood(self):
        # symmetric
        v = []
        v.append(FreeCAD.Vector(0, 0, 0.0))
        v.append(FreeCAD.Vector(1, 0, 0.5))
        v.append(FreeCAD.Vector(2, 0, 0.0))
        v.append(FreeCAD.Vector(0, 1, 0.5))
        v.append(FreeCAD.Vector(1, 1, 1.0))
        v.append(FreeCAD.Vector(2, 1, 0.5))
        v.append(FreeCAD.Vector(0, 2, 0.0))
        v.append(FreeCAD.Vector(1, 2, 0.5))
        v.append(FreeCAD.Vector(2, 2, 0.0))
        d = Mesh.polynomialFit(v)
        c = d["Coefficients"]
        # print ("Polynomial: f(x,y)=%f*x^2%+f*y^2%+f*x*y%+f*x%+f*y%+f" % (c[0],c[1],c[2],c[3],c[4],c[5]))
        for i in d["Residuals"]:
            self.assertTrue(math.fabs(i) < 0.0001, "Too high residual %f" % math.fabs(i))

    def testFitExact(self):
        # symmetric
        v = []
        v.append(FreeCAD.Vector(0, 0, 0.0))
        v.append(FreeCAD.Vector(1, 0, 0.0))
        v.append(FreeCAD.Vector(2, 0, 0.0))
        v.append(FreeCAD.Vector(0, 1, 0.0))
        v.append(FreeCAD.Vector(1, 1, 1.0))
        v.append(FreeCAD.Vector(2, 1, 0.0))
        d = Mesh.polynomialFit(v)
        c = d["Coefficients"]
        # print ("Polynomial: f(x,y)=%f*x^2%+f*y^2%+f*x*y%+f*x%+f*y%+f" % (c[0],c[1],c[2],c[3],c[4],c[5]))
        for i in d["Residuals"]:
            self.assertTrue(math.fabs(i) < 0.0001, "Too high residual %f" % math.fabs(i))

    def testFitBad(self):
        # symmetric
        v = []
        v.append(FreeCAD.Vector(0, 0, 0.0))
        v.append(FreeCAD.Vector(1, 0, 0.0))
        v.append(FreeCAD.Vector(2, 0, 0.0))
        v.append(FreeCAD.Vector(0, 1, 0.0))
        v.append(FreeCAD.Vector(1, 1, 1.0))
        v.append(FreeCAD.Vector(2, 1, 0.0))
        v.append(FreeCAD.Vector(0, 2, 0.0))
        v.append(FreeCAD.Vector(1, 2, 0.0))
        v.append(FreeCAD.Vector(2, 2, 0.0))
        d = Mesh.polynomialFit(v)
        c = d["Coefficients"]
        # print ("Polynomial: f(x,y)=%f*x^2%+f*y^2%+f*x*y%+f*x%+f*y%+f" % (c[0],c[1],c[2],c[3],c[4],c[5]))
        for i in d["Residuals"]:
            self.assertFalse(math.fabs(i) < 0.0001, "Residual %f must be higher" % math.fabs(i))

    def tearDown(self):
        pass


class NastranReader(unittest.TestCase):
    def setUp(self):
        self.test_dir = join(FreeCAD.getHomePath(), "Mod", "Mesh", "App", "TestData")

    def testEightCharGRIDElement(self):
        m = Mesh.read(f"{self.test_dir}/NASTRAN_Test_GRID_CQUAD4.bdf")
        self.assertEqual(m.CountPoints, 10)
        self.assertEqual(m.CountFacets, 8)  # Quads split into two triangles

    def testDelimitedGRIDElement(self):
        m = Mesh.read(f"{self.test_dir}/NASTRAN_Test_Delimited_GRID_CQUAD4.bdf")
        self.assertEqual(m.CountPoints, 10)
        self.assertEqual(m.CountFacets, 8)  # Quads split into two triangles

    def testSixteenCharGRIDElement(self):
        m = Mesh.read(f"{self.test_dir}/NASTRAN_Test_GRIDSTAR_CQUAD4.bdf")
        self.assertEqual(m.CountPoints, 4)
        self.assertEqual(m.CountFacets, 2)  # Quads split into two triangles

    def testCTRIA3Element(self):
        m = Mesh.read(f"{self.test_dir}/NASTRAN_Test_GRID_CTRIA3.bdf")
        self.assertEqual(m.CountPoints, 3)
        self.assertEqual(m.CountFacets, 1)

    def tearDown(self):
        pass


class MeshSubElement(unittest.TestCase):
    def setUp(self):
        self.mesh = Mesh.createBox(1.0, 1.0, 1.0)

    def testCenterOfGravity(self):
        c = self.mesh.CenterOfGravity
        self.assertEqual(c, Base.Vector(0.0, 0.0, 0.0))

    def testSubElements(self):
        types = self.mesh.getElementTypes()
        self.assertIn("Mesh", types)
        self.assertIn("Segment", types)

    def testCountSubElements(self):
        self.assertEqual(self.mesh.countSubElements("Mesh"), 1)
        self.assertEqual(self.mesh.countSubElements("Segment"), 0)

    def testFacesFromSubElement(self):
        element = self.mesh.getFacesFromSubElement("Mesh", 0)
        self.assertIsInstance(element, tuple)
        self.assertEqual(len(element), 2)
        self.assertEqual(len(element[0]), 8)
        self.assertEqual(len(element[1]), 12)

    def testSegmentSubElement(self):
        self.mesh.addSegment([0, 2, 4, 6, 8])
        self.assertEqual(self.mesh.countSegments(), 1)
        self.assertEqual(self.mesh.countSubElements("Segment"), 1)
        element = self.mesh.getFacesFromSubElement("Segment", 0)
        self.assertIsInstance(element, tuple)
        self.assertEqual(len(element), 2)
        self.assertEqual(len(element[0]), 7)
        self.assertEqual(len(element[1]), 5)
        segment = self.mesh.meshFromSegment(self.mesh.getSegment(0))
        self.assertEqual(segment.CountPoints, 7)
        self.assertEqual(segment.CountFacets, 5)

    def tearDown(self):
        pass


class MeshProperty(unittest.TestCase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("MeshTest")

    def tearDown(self):
        FreeCAD.closeDocument(self.doc.Name)

    def testMaterial(self):
        mesh = self.doc.addObject("Mesh::Feature", "Sphere")
        mesh.Mesh = Mesh.createBox(1.0, 1.0, 1.0)
        len1 = int(mesh.Mesh.CountFacets / 2)
        len2 = int(mesh.Mesh.CountFacets - len1)
        material = {"transparency": [0.2] * len1 + [0.8] * len2}
        material["binding"] = MeshEnums.Binding.PER_FACE
        material["ambientColor"] = [(1, 0, 0)] * (len1 + len2)
        material["diffuseColor"] = [(0, 1, 0)] * (len1 + len2)
        material["specularColor"] = [(0, 0, 1)] * (len1 + len2)
        material["emissiveColor"] = [(1, 1, 1)] * (len1 + len2)
        material["shininess"] = [0.3] * (len1 + len2)

        mesh.addProperty("Mesh::PropertyMaterial", "Material")
        mesh.Material = material

        TempPath = tempfile.gettempdir()
        SaveName = TempPath + os.sep + "mesh_with_material.FCStd"
        self.doc.saveAs(SaveName)
        FreeCAD.closeDocument(self.doc.Name)

        self.doc = FreeCAD.openDocument(SaveName)
        mesh2 = self.doc.Sphere
        material2 = mesh2.Material

        self.assertEqual(int(material2["binding"]), int(MeshEnums.Binding.PER_FACE))
        self.assertEqual(len(material2["ambientColor"]), len1 + len2)
        self.assertEqual(len(material2["diffuseColor"]), len1 + len2)
        self.assertEqual(len(material2["specularColor"]), len1 + len2)
        self.assertEqual(len(material2["emissiveColor"]), len1 + len2)
        self.assertEqual(len(material2["shininess"]), len1 + len2)
        self.assertEqual(len(material2["transparency"]), len1 + len2)
